import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic


class SkinFriction(om.ImplicitComponent):
    """
    Computes skin friction coefficient using the Sommer and Short T Prime method as used
    in FLOPS AERSCL.

    The fixed-point iteration scheme has been replaced with Newton's method, which can
    converge the equations for multiple mach numbers and characteristic lengths
    simultaneously.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.CONLOG = 2.302585
        self.TAW = 1.0
        self.sea_level_pressure = 14.6959 * 144  # psi -> psf

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=False,
                                                atol=1e-12, rtol=1e-12)
        self.linear_solver = om.DirectSolver()
        self.nonlinear_solver.options['iprint'] = -1
        self.linear_solver.options['iprint'] = -1

    def initialize(self):
        """
        Declare options.
        """
        self.options.declare(
            'num_nodes', types=int, default=1,
            desc='The number of points at which the cross product is computed.')
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nn = self.options['num_nodes']
        aviary_options: AviaryValues = self.options['aviary_options']
        zero_count = (0, None)
        num_tails, _ = aviary_options.get_item(
            Aircraft.VerticalTail.NUM_TAILS, zero_count)
        num_fuselages, _ = aviary_options.get_item(
            Aircraft.Fuselage.NUM_FUSELAGES, zero_count)
        # TODO does not used vectorized multi-engine. Temp using single engine
        num_engines, _ = aviary_options.get_item(
            Aircraft.Engine.NUM_ENGINES, zero_count)
        self.nc = nc = 2 + num_tails + num_fuselages + int(sum(num_engines))

        # Simulation inputs
        self.add_input(Dynamic.Mission.TEMPERATURE, np.ones(nn), units='degR')
        self.add_input(Dynamic.Mission.STATIC_PRESSURE, np.ones(nn), units='lbf/ft**2')
        self.add_input(Dynamic.Mission.MACH, np.ones(nn), units='unitless')

        # Aero subsystem inputs
        self.add_input('characteristic_lengths', np.ones(nc), units='ft')

        self.add_output('cf_iter', np.ones((nn, nc)), units='unitless')
        self.add_output('skin_friction_coeff', np.ones((nn, nc)), units='unitless')
        self.add_output('Re', np.ones((nn, nc)), units='unitless', res_ref=1e6)
        self.add_output('wall_temp', np.ones((nn, nc)), units='degR')

    def setup_partials(self):
        nn = self.options["num_nodes"]
        nc = self.nc
        n = nn * nc

        row_col = np.arange(n)
        self.declare_partials('Re', 'Re', rows=row_col, cols=row_col, val=1.0)
        self.declare_partials(
            'skin_friction_coeff', 'skin_friction_coeff',
            rows=row_col, cols=row_col, val=1.0)

        self.declare_partials(
            'cf_iter', ['wall_temp', 'cf_iter'], rows=row_col, cols=row_col)
        self.declare_partials(
            'wall_temp', ['wall_temp', 'cf_iter'], rows=row_col, cols=row_col)
        self.declare_partials(
            'skin_friction_coeff', ['wall_temp', 'cf_iter'], rows=row_col, cols=row_col)

        col = np.arange(nn)
        cols = np.repeat(col, nc)
        self.declare_partials(
            'cf_iter', [Dynamic.Mission.TEMPERATURE, Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.MACH], rows=row_col, cols=cols)
        self.declare_partials(
            'wall_temp', [Dynamic.Mission.TEMPERATURE, Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.MACH], rows=row_col, cols=cols)
        self.declare_partials(
            'Re', [Dynamic.Mission.TEMPERATURE, Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.MACH], rows=row_col, cols=cols)
        self.declare_partials(
            'skin_friction_coeff', [Dynamic.Mission.TEMPERATURE,
                                    Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.MACH],
            rows=row_col, cols=cols)

        col = np.arange(nc)
        cols = np.tile(col, nn)
        self.declare_partials('Re', 'characteristic_lengths', rows=row_col, cols=cols)
        self.declare_partials(
            'cf_iter', 'characteristic_lengths', rows=row_col, cols=cols)

    def guess_nonlinear(self, inputs, outputs, resids):
        nn = self.options["num_nodes"]
        nc = self.nc

        rankine, pressure, mach, length = inputs.values()

        Pratio = pressure / self.sea_level_pressure
        kelvin = rankine / 1.8
        RE = 1.479301E9 * Pratio * (kelvin + 110.4) / kelvin ** 2

        # REYNOLDS NUMBER
        reynolds_num = np.einsum('i,i,j->ij', RE, mach, length)
        outputs['Re'] = reynolds_num

        # Initial guess for WALL TEMPERATURE
        wall_temp = (1.0 + 0.176 * mach * mach) * rankine
        wall_temp = np.tile(wall_temp, nc).reshape((nc, nn)).T
        outputs['wall_temp'] = wall_temp
        self.TAW = wall_temp

        # INITIAL GUESS AT SKIN FRICTION COEFFICIENT
        outputs['cf_iter'] = (0.242 / (np.log(reynolds_num * 0.0015) / self.CONLOG)) ** 2

    def apply_nonlinear(self, inputs, outputs, residuals):
        T, pressure, mach, length = inputs.values()
        cf = outputs['cf_iter']
        wall_temp = outputs['wall_temp']
        nc = self.nc

        Pratio = pressure / self.sea_level_pressure
        kelvin = T / 1.8
        RE = 1.479301E9 * Pratio * (kelvin + 110.4) / kelvin ** 2

        # SUTHERLAND'S CONSTANT IS 198.72 DEG R FROM 1962 ON
        suth_const = T + 198.72
        E = 0.80

        # COMBINED CONSTANT INCLUDING 1/RHO
        combined_const = 4.593153E-6 * E * suth_const / (RE * mach * T ** 1.5)

        # REYNOLDS NUMBER
        reynolds_num = np.einsum('i,i,j->ij', RE, mach, length)
        residuals['Re'] = outputs['Re'] - reynolds_num

        # WALL TEMPERATURE RATIO
        wall_temp_ratio = 1.0 + 0.45 * (np.einsum('ij,i->ij', wall_temp,  1.0 / T) - 1.0)
        fact = 0.035 * mach * mach
        for i in range(nc):
            wall_temp_ratio[:, i] += fact

        CFL = cf / (1.0 + 3.59 * np.sqrt(cf) * wall_temp_ratio)

        prod = np.einsum('j,jk->jk', combined_const, wall_temp ** 3)
        residuals['wall_temp'] = (
            self.TAW / (1.0 + prod / CFL) + wall_temp) * 0.5 - wall_temp

        RP = (
            reynolds_num * (np.einsum('ij,i->ij', wall_temp_ratio, T) + 198.72)
            / np.einsum('i,ij->ij', suth_const, wall_temp_ratio ** 2.5))

        residuals['cf_iter'] = (0.242 * self.CONLOG / np.log(RP * cf)) ** 2 - cf

        residuals['skin_friction_coeff'] = \
            outputs['skin_friction_coeff'] - outputs['cf_iter'] / wall_temp_ratio

    def linearize(self, inputs, outputs, partials):
        nn = self.options["num_nodes"]
        nc = self.nc

        T, pressure, mach, length = inputs.values()
        cf = outputs['cf_iter']
        wall_temp = outputs['wall_temp']

        Pratio = pressure / self.sea_level_pressure
        kelvin = T / 1.8
        RE = 1.479301E9 * Pratio * (kelvin + 110.4) / kelvin ** 2

        dRE_dp = 1.479301E9 * (kelvin + 110.4) / (self.sea_level_pressure * kelvin ** 2)
        dRE_dT = -1.479301E9 / 1.8 * (
            Pratio * (1.0 / kelvin ** 2 + 2.0 * 110.4 / kelvin ** 3))

        reynolds_num = np.einsum('i,i,j->ij', RE, mach, length)
        dreyn_dp = np.einsum('i,i,j->ij', dRE_dp, mach, length)
        dreyn_dT = np.einsum('i,i,j->ij', dRE_dT, mach, length)
        dreyn_dmach = np.einsum('i,j->ij', RE, length)
        dreyn_dlen = np.tile(RE * mach, nc).reshape((nc, nn)).T

        partials['Re', Dynamic.Mission.STATIC_PRESSURE] = -dreyn_dp.ravel()
        partials['Re', Dynamic.Mission.TEMPERATURE] = -dreyn_dT.ravel()
        partials['Re', Dynamic.Mission.MACH] = -dreyn_dmach.ravel()
        partials['Re', 'characteristic_lengths'] = -dreyn_dlen.ravel()

        suth_const = T + 198.72
        E = 0.80
        combined_const = 4.593153E-6 * E * suth_const / (RE * mach * T ** 1.5)
        dcomb_dRE = -4.593153E-6 * E * suth_const / (RE * RE * mach * T ** 1.5)
        dcomb_dmach = -4.593153E-6 * E * suth_const / (RE * mach * mach * T ** 1.5)
        dcomb_dT = 4.593153E-6 * E * (1.0 / T ** 1.5 - 1.5 * suth_const / T ** 2.5) / (
            RE * mach) + dcomb_dRE * dRE_dT
        dcomb_dp = dcomb_dRE * dRE_dp

        wall_temp_ratio = 1.0 + 0.45 * (np.einsum('ij,i->ij', wall_temp,  1.0 / T) - 1.0)
        fact = 0.035 * mach * mach
        for i in range(nc):
            wall_temp_ratio[:, i] += fact
        dwtr_dmach = 0.07 * mach
        dwtr_dT = -0.45 * np.einsum('ij,i->ij', wall_temp, 1.0 / T ** 2)
        dwtr_dwt = 0.45 / T

        den = 1.0 + 3.59 * np.sqrt(cf) * wall_temp_ratio
        CFL = cf / den
        dCFL_dcf = \
            1.0 / den - cf * 3.59 * wall_temp_ratio * 0.5 / (np.sqrt(cf) * den ** 2)
        dCFL_dwtr = - cf * 3.59 * np.sqrt(cf) / den ** 2
        dCFL_dmach = np.einsum('ij,i->ij', dCFL_dwtr, dwtr_dmach)
        dCFL_dT = dCFL_dwtr * dwtr_dT
        dCFL_dwt = np.einsum('ij,i->ij', dCFL_dwtr, dwtr_dwt)

        term = np.einsum('i,ij->ij', combined_const, wall_temp ** 3)
        den = 1.0 + term / CFL
        dreswt_dcomb = -0.5 * self.TAW * wall_temp ** 3 / (CFL * den ** 2)
        dreswt_dCFL = 0.5 * self.TAW * term / (CFL * den) ** 2
        dreswt_dwt = (
            -0.5 - 1.5 * self.TAW * np.einsum('i,ij->ij', combined_const, wall_temp ** 2)
            / (CFL * den ** 2))

        partials['wall_temp', Dynamic.Mission.STATIC_PRESSURE] = (
            np.einsum('ij,i->ij', dreswt_dcomb, dcomb_dp)).ravel()
        partials['wall_temp', Dynamic.Mission.TEMPERATURE] = (
            np.einsum('ij,i->ij', dreswt_dcomb, dcomb_dT)
            + dreswt_dCFL * dCFL_dT).ravel()
        partials['wall_temp', Dynamic.Mission.MACH] = (
            np.einsum('ij,i->ij', dreswt_dcomb, dcomb_dmach)
            + dreswt_dCFL * dCFL_dmach).ravel()
        partials['wall_temp', 'wall_temp'] = (
            dreswt_dCFL * dCFL_dwt + dreswt_dwt).ravel()
        partials['wall_temp', 'cf_iter'] = (dreswt_dCFL * dCFL_dcf).ravel()

        den = 1.0 / np.einsum('i,ij->ij', suth_const, wall_temp_ratio ** 2.5)
        num = np.einsum('ij,i->ij', wall_temp_ratio, T) + 198.72
        RP = reynolds_num * num * den
        dRP_dreyn = num * den
        term = np.einsum('i,ij->ij', T, 1.0 * den)
        dRP_dwtr = reynolds_num * (
            term - num * 2.5
            * np.einsum('i,ij->ij', suth_const, wall_temp_ratio ** 1.5) * den ** 2)
        dRP_dT = reynolds_num * (
            wall_temp_ratio * den - num * wall_temp_ratio ** 2.5 * den ** 2)
        dRP_dp = dRP_dreyn * dreyn_dp
        dRP_dT = dRP_dreyn * dreyn_dT + dRP_dwtr * dwtr_dT + dRP_dT
        dRP_dmach = dRP_dreyn * dreyn_dmach + np.einsum('ij,i->ij', dRP_dwtr, dwtr_dmach)
        dRP_dlen = dRP_dreyn * dreyn_dlen
        dRP_dwt = np.einsum('ij,i->ij', dRP_dwtr, dwtr_dwt)

        fact = (0.242 * self.CONLOG) ** 2
        drescf_dRP = -2.0 * fact / (RP * np.log(RP * cf) ** 3)
        drescf_dcf = -2.0 * fact / (cf * np.log(RP * cf) ** 3) - 1.0

        partials['cf_iter', Dynamic.Mission.STATIC_PRESSURE] = (
            drescf_dRP * dRP_dp).ravel()
        partials['cf_iter', Dynamic.Mission.TEMPERATURE] = (drescf_dRP * dRP_dT).ravel()
        partials['cf_iter', Dynamic.Mission.MACH] = (drescf_dRP * dRP_dmach).ravel()
        partials['cf_iter', 'characteristic_lengths'] = (drescf_dRP * dRP_dlen).ravel()
        partials['cf_iter', 'wall_temp'] = (drescf_dRP * dRP_dwt).ravel()
        partials['cf_iter', 'cf_iter'] = drescf_dcf.ravel()

        dskf_dwtr = outputs['cf_iter'] / wall_temp_ratio ** 2

        partials['skin_friction_coeff', Dynamic.Mission.TEMPERATURE] = (
            dskf_dwtr * dwtr_dT).ravel()
        partials['skin_friction_coeff', Dynamic.Mission.MACH] = np.einsum(
            'ij,i->ij', dskf_dwtr, dwtr_dmach).ravel()
        partials['skin_friction_coeff', 'wall_temp'] = np.einsum(
            'ij,i->ij', dskf_dwtr, dwtr_dwt).ravel()
        partials['skin_friction_coeff', 'cf_iter'] = (- 1.0 / wall_temp_ratio).ravel()



import unittest

from openmdao.utils.assert_utils import assert_near_equal

from aviary.examples.variable_meta_data_extension import ExtendedMetaData
from aviary.examples.variables_extension import Aircraft
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.functions import set_value
from aviary.utils.functions import get_path


class MetaDataExtensionTest(unittest.TestCase):
    def test_set_value_metadata_extension(self):
        option_defaults = AviaryValues()

        filename = get_path(
            'models/engines/turbofan_23k_1.deck')
        option_defaults.set_val(Aircraft.Engine.DATA_FILE, filename)
        option_defaults.set_val(Aircraft.Engine.NUM_ENGINES, val=[
                                2], units='unitless', meta_data=ExtendedMetaData)
        option_defaults.set_val(Aircraft.Wing.AERO_CENTER, val=5, units='ft', meta_data=ExtendedMetaData
                                )

        option_defaults = set_value(Aircraft.Wing.AERO_CENTER, [
                                    1.0], option_defaults, 'ft', ExtendedMetaData)

        check_val = option_defaults.get_val(Aircraft.Wing.AERO_CENTER, units='inch')

        assert_near_equal(check_val, 12, 1e-9)


if __name__ == '__main__':
    unittest.main()



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import \
    distributed_engine_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class TransportFuselageMass(om.ExplicitComponent):
    """
    Computes the mass of the fuselage for transports. The methodology
    is based on the FLOPS weight equations, modified to output mass instead
    of weight.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=0.0)

        add_aviary_output(self, Aircraft.Fuselage.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Fuselage.MASS, "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        length = inputs[Aircraft.Fuselage.LENGTH]
        scaler = inputs[Aircraft.Fuselage.MASS_SCALER]
        avg_diameter = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        num_fuse = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)
        military_cargo = aviary_options.get_val(Aircraft.Fuselage.MILITARY_CARGO_FLOOR)

        mil_factor = 1.38 if military_cargo else 1.0

        outputs[Aircraft.Fuselage.MASS] = (
            scaler * 1.35 * (avg_diameter * length) ** 1.28 * (1.0 + 0.05 * num_fuse_eng_fact) *
            mil_factor * num_fuse / GRAV_ENGLISH_LBM
        )

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        length = inputs[Aircraft.Fuselage.LENGTH]
        scaler = inputs[Aircraft.Fuselage.MASS_SCALER]
        avg_diameter = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        num_fuse = aviary_options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)
        military_cargo = aviary_options.get_val(Aircraft.Fuselage.MILITARY_CARGO_FLOOR)

        # avg_diameter = (max_height + max_width) / 2.
        avg_diameter_exp = avg_diameter ** 1.28
        length_exp = length ** 1.28
        height_width_exp = (avg_diameter)**0.28
        mil_factor = 1.38 if military_cargo else 1.0
        addtl_factor = (1.0 + 0.05 * num_fuse_eng_fact) * mil_factor * num_fuse

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.MASS_SCALER] = 1.35 * \
            avg_diameter_exp * length_exp * addtl_factor / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.LENGTH] = scaler * \
            1.728 * avg_diameter_exp * length ** 0.28 * addtl_factor / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.AVG_DIAMETER] = scaler * \
            1.728 * length_exp * height_width_exp * addtl_factor / GRAV_ENGLISH_LBM


class AltFuselageMass(om.ExplicitComponent):
    """
    Computes the mass of the fuselage using the alternate method. The methodology
    is based on the FLOPS weight equations, modified to output mass instead
    of weight.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.MASS_SCALER, 1.0)

        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, 0.0)

        add_aviary_input(self, Aircraft.Fuselage.MAX_HEIGHT, 0.0)

        add_aviary_input(self, Aircraft.Fuselage.MAX_WIDTH, 0.0)

        add_aviary_output(self, Aircraft.Fuselage.MASS, 1.0)

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):
        mass_scaler = inputs[Aircraft.Fuselage.MASS_SCALER]
        fuse_wetted_area = inputs[Aircraft.Fuselage.WETTED_AREA]
        fuse_height = inputs[Aircraft.Fuselage.MAX_HEIGHT]
        fuse_width = inputs[Aircraft.Fuselage.MAX_WIDTH]

        outputs[Aircraft.Fuselage.MASS] = (
            3.939 * fuse_wetted_area /
            (fuse_height / fuse_width)**0.221 * mass_scaler / GRAV_ENGLISH_LBM
        )

    def compute_partials(self, inputs, J):
        mass_scaler = inputs[Aircraft.Fuselage.MASS_SCALER]
        fuse_wetted_area = inputs[Aircraft.Fuselage.WETTED_AREA]
        fuse_height = inputs[Aircraft.Fuselage.MAX_HEIGHT]
        fuse_width = inputs[Aircraft.Fuselage.MAX_WIDTH]
        fuse_height_fact = fuse_height ** 0.221
        fuse_width_fact = fuse_width ** 0.221
        total_fact = fuse_height_fact / fuse_width_fact

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.MASS_SCALER] = (
            3.939 * fuse_wetted_area / total_fact / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.WETTED_AREA] = (
            3.939 / total_fact * mass_scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.MAX_HEIGHT] = (
            -0.870519 * fuse_wetted_area * fuse_width_fact
            * fuse_height**-1.221 * mass_scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.MAX_WIDTH] = (
            0.870519 * fuse_wetted_area * fuse_width**-0.779 /
            fuse_height_fact * mass_scaler / GRAV_ENGLISH_LBM
        )



import numpy as np
import openmdao.api as om

from aviary import constants
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Dynamic


class FlightConditions(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare(
            "input_speed_type",
            default=SpeedType.TAS,
            types=SpeedType,
            desc="tells whether the input airspeed is equivalent airspeed, true airspeed, or mach number",
        )

    def setup(self):
        nn = self.options["num_nodes"]
        in_type = self.options["input_speed_type"]
        arange = np.arange(self.options["num_nodes"])

        self.add_input(
            "rho",
            val=np.zeros(nn),
            units="slug/ft**3",
            desc="density of air",
        )
        self.add_input(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=np.zeros(nn),
            units="ft/s",
            desc="speed of sound",
        )

        self.add_output(
            Dynamic.Mission.DYNAMIC_PRESSURE,
            val=np.zeros(nn),
            units="lbf/ft**2",
            desc="dynamic pressure",
        )

        if in_type is SpeedType.TAS:
            self.add_input(
                "TAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="true air speed",
            )
            self.add_output(
                "EAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="equivalent air speed",
            )
            self.add_output(
                Dynamic.Mission.MACH,
                val=np.zeros(nn),
                units="unitless",
                desc="mach number",
            )

            self.declare_partials(Dynamic.Mission.DYNAMIC_PRESSURE, [
                                  "rho", "TAS"], rows=arange, cols=arange)
            self.declare_partials(Dynamic.Mission.MACH, [
                                  Dynamic.Mission.SPEED_OF_SOUND, "TAS"], rows=arange, cols=arange)
            self.declare_partials("EAS", ["TAS", "rho"], rows=arange, cols=arange)
        elif in_type is SpeedType.EAS:
            self.add_input(
                "EAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="equivalent air speed at",
            )
            self.add_output(
                "TAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="true air speed",
            )
            self.add_output(
                Dynamic.Mission.MACH,
                val=np.zeros(nn),
                units="unitless",
                desc="mach number",
            )

            self.declare_partials(Dynamic.Mission.DYNAMIC_PRESSURE, [
                                  "rho", "EAS"], rows=arange, cols=arange)
            self.declare_partials(
                Dynamic.Mission.MACH, [Dynamic.Mission.SPEED_OF_SOUND, "EAS", "rho"], rows=arange, cols=arange
            )
            self.declare_partials("TAS", ["rho", "EAS"], rows=arange, cols=arange)
        elif in_type is SpeedType.MACH:
            self.add_input(
                Dynamic.Mission.MACH,
                val=np.zeros(nn),
                units="unitless",
                desc="mach number",
            )
            self.add_output(
                "EAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="equivalent air speed",
            )
            self.add_output(
                "TAS",
                val=np.zeros(nn),
                units="ft/s",
                desc="true air speed",
            )

            self.declare_partials(
                Dynamic.Mission.DYNAMIC_PRESSURE, [Dynamic.Mission.SPEED_OF_SOUND, Dynamic.Mission.MACH, "rho"], rows=arange, cols=arange)
            self.declare_partials(
                "TAS", [Dynamic.Mission.SPEED_OF_SOUND, Dynamic.Mission.MACH], rows=arange, cols=arange)
            self.declare_partials(
                "EAS", [Dynamic.Mission.SPEED_OF_SOUND, Dynamic.Mission.MACH, "rho"], rows=arange, cols=arange
            )

    def compute(self, inputs, outputs):

        in_type = self.options["input_speed_type"]

        rho = inputs["rho"]
        sos = inputs[Dynamic.Mission.SPEED_OF_SOUND]

        if in_type is SpeedType.TAS:
            TAS = inputs["TAS"]
            outputs[Dynamic.Mission.MACH] = mach = TAS / sos
            outputs["EAS"] = EAS = TAS * (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            outputs[Dynamic.Mission.DYNAMIC_PRESSURE] = q = 0.5 * rho * TAS**2

        elif in_type is SpeedType.EAS:
            EAS = inputs["EAS"]
            outputs["TAS"] = TAS = EAS / (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            outputs[Dynamic.Mission.MACH] = mach = TAS / sos
            outputs[Dynamic.Mission.DYNAMIC_PRESSURE] = q = 0.5 * \
                EAS**2 * constants.RHO_SEA_LEVEL_ENGLISH

        elif in_type is SpeedType.MACH:
            mach = inputs[Dynamic.Mission.MACH]
            outputs["TAS"] = TAS = sos * mach
            outputs["EAS"] = EAS = TAS * (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            outputs[Dynamic.Mission.DYNAMIC_PRESSURE] = 0.5 * rho * sos**2 * mach**2

    def compute_partials(self, inputs, J):
        in_type = self.options["input_speed_type"]

        rho = inputs["rho"]
        sos = inputs[Dynamic.Mission.SPEED_OF_SOUND]

        if in_type is SpeedType.TAS:
            TAS = inputs["TAS"]

            J[Dynamic.Mission.DYNAMIC_PRESSURE, "TAS"] = rho * TAS
            J[Dynamic.Mission.DYNAMIC_PRESSURE, "rho"] = 0.5 * TAS**2

            J[Dynamic.Mission.MACH, "TAS"] = 1 / sos
            J[Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND] = -TAS / sos**2

            J["EAS", "TAS"] = (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            J["EAS", "rho"] = (
                TAS * 0.5 * (rho ** (-0.5) / constants.RHO_SEA_LEVEL_ENGLISH**0.5)
            )

        elif in_type is SpeedType.EAS:
            EAS = inputs["EAS"]
            TAS = EAS / (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5

            dTAS_dRho = -0.5 * EAS * constants.RHO_SEA_LEVEL_ENGLISH**0.5 / rho**1.5
            dTAS_dEAS = 1 / (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5

            J[Dynamic.Mission.DYNAMIC_PRESSURE, "EAS"] = EAS * \
                constants.RHO_SEA_LEVEL_ENGLISH
            J[Dynamic.Mission.MACH, "EAS"] = dTAS_dEAS / sos
            J[Dynamic.Mission.MACH, "rho"] = dTAS_dRho / sos
            J[Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND] = -TAS / sos**2
            J["TAS", "rho"] = dTAS_dRho
            J["TAS", "EAS"] = dTAS_dEAS

        elif in_type is SpeedType.MACH:
            mach = inputs[Dynamic.Mission.MACH]
            TAS = sos * mach

            J[Dynamic.Mission.DYNAMIC_PRESSURE,
                Dynamic.Mission.SPEED_OF_SOUND] = rho * sos * mach**2
            J[Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.MACH] = rho * sos**2 * mach
            J[Dynamic.Mission.DYNAMIC_PRESSURE, "rho"] = 0.5 * sos**2 * mach**2
            J["TAS", Dynamic.Mission.SPEED_OF_SOUND] = mach
            J["TAS", Dynamic.Mission.MACH] = sos
            J["EAS", Dynamic.Mission.SPEED_OF_SOUND] = mach * \
                (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            J["EAS", Dynamic.Mission.MACH] = sos * \
                (rho / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5
            J["EAS", "rho"] = (
                TAS * (1 / constants.RHO_SEA_LEVEL_ENGLISH) ** 0.5 * 0.5 * rho ** (-0.5)
            )



from openmdao.core.system import System

from aviary.utils.aviary_values import AviaryValues, get_keys


def list_options(model: System, aviary_keys: list = None):
    """
    Lists all option values in the provided model. All top-level option values will
    be listed, and items in model.options['aviary_options'] will also be listed.
    A list of keys may be provided to limit the list of items in aviary_options.

    Parameters
    ----------
    model : System
        A model.
    aviary_keys: iter of str
        List of aviary_options keys whose values will be looked up and
        listed in the options printout. If None, all items in
        model.options['aviary_options'] will be listed.
    """
    print('\nOptions:\n')
    for subsystem in model.system_iter():
        if subsystem.name == '_auto_ivc':
            continue
        print(subsystem.name)
        for (key, obj) in subsystem.options.items():
            if isinstance(obj, AviaryValues):
                aviary_options = obj
                print('  aviary_options:')
                if isinstance(aviary_keys, list):
                    keys = aviary_keys
                else:
                    keys = get_keys(aviary_options)
                for key in keys:
                    (val, units) = aviary_options.get_item(key)
                    if units == 'unitless':
                        print(f'    {key} = {val}')
                    else:
                        print(f'    {key} = {val} {units}')
            else:
                print(f'  {key} = {str(obj)[0:80]}')
    print()



import numpy as np
import openmdao.api as om

from aviary.subsystems.mass.flops_based.distributed_prop import nacelle_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class EnginePodMass(om.ExplicitComponent):
    '''
    Calculates the mass of a single engine pod

    ASSUMPTIONS: System-level masses are estimated per engine by normalizing with
                 ratio of total engine model set's thrust to total aircraft thrust
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Electrical.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Hydraulics.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Instruments.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Nacelle.MASS, val=np.zeros(1))
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Engine.MASS, val=np.zeros(1))
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_STARTER_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Engine.THRUST_REVERSERS_MASS, val=np.zeros(1))
        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=np.zeros(1))
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST, val=0.0)

        add_aviary_output(self, Aircraft.Engine.POD_MASS, val=np.zeros(1))

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        # TODO this methodology completely ignores miscellaneous mass. There is a discrepency between this calculation
        #      and miscellaneous mass. Engine control, starter, and additional mass have a scaler applied to them, and
        #      if their calculated values are used directly this scaler is skipped
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        nacelle_count = nacelle_count_factor(num_eng)

        eng_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        total_thrust = inputs[Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST]

        # TODO this value may be needed elsewhere, if so, calculate in static
        # propulsion and pass to this component as input
        engine_thrust_ratio = (eng_thrust * num_eng) / total_thrust

        m_eng = inputs[Aircraft.Engine.MASS]
        m_thr = inputs[Aircraft.Engine.THRUST_REVERSERS_MASS]
        m_start = inputs[Aircraft.Propulsion.TOTAL_STARTER_MASS]
        m_ctrl = inputs[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS]
        m_inst = inputs[Aircraft.Instruments.MASS]
        m_elec = inputs[Aircraft.Electrical.MASS]
        m_hyd = inputs[Aircraft.Hydraulics.MASS]
        m_fsys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS]
        m_nac = inputs[Aircraft.Nacelle.MASS]

        # calculate nacelle content mass for engine set
        nacelle_content_mass = (
            (num_eng * m_eng)
            + m_thr
            + engine_thrust_ratio
            * (
                m_start
                + .25 * (m_ctrl + m_fsys)
                + .13 * (m_elec + m_hyd)
                + .11 * m_inst
            )
        )

        pod_mass = np.array([])
        for i in range(len(aviary_options.get_val('engine_models'))):
            pod_mass = np.append(pod_mass,
                                 nacelle_content_mass[i] / max(1,
                                                               num_eng[i]) + m_nac[i] / max(1,
                                                                                            nacelle_count[i]))

        # calculate engine pod mass for single engine
        outputs[Aircraft.Engine.POD_MASS] = pod_mass

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        count_factor = nacelle_count_factor(num_eng)

        m_start = inputs[Aircraft.Propulsion.TOTAL_STARTER_MASS]
        m_ctrl = inputs[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS]
        m_inst = inputs[Aircraft.Instruments.MASS]
        m_elec = inputs[Aircraft.Electrical.MASS]
        m_hyd = inputs[Aircraft.Hydraulics.MASS]
        m_fsys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS]
        m_nac = inputs[Aircraft.Nacelle.MASS]

        eng_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        total_thrust = inputs[Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST]
        # TODO this value may be needed elsewhere, if so, calculate in static
        # propulsion and pass to this component as input
        ratio = eng_thrust * num_eng / total_thrust

        fact1 = np.array([1.0 / max(1, eng) for eng in num_eng])
        fact2 = np.array(1.0 / max(1, count_factor))
        nac_fact = (m_start + 0.25 * (m_ctrl + m_fsys) +
                    0.13 * (m_elec + m_hyd) + 0.11 * m_inst)

        partials[Aircraft.Engine.POD_MASS, Aircraft.Engine.MASS] = 1.0

        partials[Aircraft.Engine.POD_MASS, Aircraft.Engine.THRUST_REVERSERS_MASS] = \
            fact1

        partials[Aircraft.Engine.POD_MASS,
                 Aircraft.Propulsion.TOTAL_STARTER_MASS] = fact1 * ratio

        partials[Aircraft.Engine.POD_MASS,
                 Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS] = .25 * fact1 * ratio

        partials[Aircraft.Engine.POD_MASS, Aircraft.Instruments.MASS] = \
            .11 * fact1 * ratio

        partials[Aircraft.Engine.POD_MASS, Aircraft.Electrical.MASS] = \
            .13 * fact1 * ratio

        partials[Aircraft.Engine.POD_MASS, Aircraft.Hydraulics.MASS] = \
            .13 * fact1 * ratio

        partials[Aircraft.Engine.POD_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS] = \
            .25 * fact1 * ratio

        partials[Aircraft.Engine.POD_MASS, Aircraft.Nacelle.MASS] = \
            fact2

        partials[Aircraft.Engine.POD_MASS, Aircraft.Engine.SCALED_SLS_THRUST] = \
            (num_eng * nac_fact * fact1) / total_thrust

        partials[
            Aircraft.Engine.POD_MASS,
            Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST] = \
            -(num_eng * eng_thrust * nac_fact * fact1) / (total_thrust ** 2)



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.propulsion.propulsion_premission import (
    PropulsionPreMission, PropulsionSum)
from aviary.utils.aviary_values import AviaryValues
from aviary.validation_cases.validation_tests import get_flops_inputs
from aviary.variable_info.variables import Aircraft


class PropulsionPreMissionTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    def test_case(self):
        aviary_values = get_flops_inputs('LargeSingleAisle2FLOPS')
        options = aviary_values

        self.prob.model = PropulsionPreMission(aviary_options=options)

        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val(Aircraft.Engine.SCALED_SLS_THRUST, options.get_val(
            Aircraft.Engine.SCALED_SLS_THRUST, units='lbf'))

        self.prob.run_model()

        sls_thrust = self.prob.get_val(Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST)

        expected_sls_thrust = np.array([54602.])

        assert_near_equal(sls_thrust, expected_sls_thrust, tolerance=1e-10)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)

    def test_propulsion_sum(self):
        options = AviaryValues()
        options.set_val(Aircraft.Engine.NUM_ENGINES, np.array([1, 2, 5]))
        # it doesn't matter what goes in engine models, as long as it is length 3
        options.set_val('engine_models', [1, 1, 1])
        self.prob.model = om.Group()
        self.prob.model.add_subsystem('propsum',
                                      PropulsionSum(aviary_options=options),
                                      promotes=['*'])

        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val(Aircraft.Engine.SCALED_SLS_THRUST,
                          np.array([1000, 3000, 13200]), units='lbf')

        self.prob.run_model()

        total_thrust = self.prob.get_val(
            Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST, units='lbf')

        expected_thrust = 73000.0

        assert_near_equal(total_thrust, expected_thrust, tolerance=1e-12)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


def sig(x):
    return 1 / (1 + np.exp(-100 * x))


def dsig(x):
    return 100 * np.exp(-100 * x) / (np.exp(-100 * x) + 1) ** 2


def dquotient(u, v, du, dv):
    return (du * v - u * dv) / v**2


class LoadSpeeds(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Design.MAX_STRUCTURAL_SPEED,
                         val=200, units="mi/h")

        if self.options["aviary_options"].get_val(Aircraft.Design.PART25_STRUCTURAL_CATEGORY, units='unitless') < 3:

            add_aviary_input(self, Aircraft.Wing.LOADING, val=128)

        self.add_output(
            "max_airspeed",
            val=0,
            units="kn",
            desc="VM0: maximum operating equivalent airspeed",
        )
        self.add_output(
            "vel_c",
            val=0,
            units="kn",
            desc="VGC: Velocity used in Gust Load Factor calculation at cruise conditions.\
                This is Minimum Design Cruise Speed for Part 23 aircraft and VM0 for Part 25 aircraft",
        )
        self.add_output(
            "max_maneuver_factor",
            val=0,
            units="unitless",
            desc="EMLF: maximum maneuver load factor, units are in g`s",
        )
        self.add_output("min_dive_vel", val=0, units="kn", desc="VDMIN: dive velocity")

        self.declare_partials("*", "*", val=0.0)

    def compute(self, inputs, outputs):

        max_struct_speed_mph = inputs[Aircraft.Design.MAX_STRUCTURAL_SPEED]

        CATD = self.options["aviary_options"].get_val(
            Aircraft.Design.PART25_STRUCTURAL_CATEGORY, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')
        WGS_greater_than_20_flag = self.options["aviary_options"].get_val(
            Aircraft.Wing.LOADING_ABOVE_20, units='unitless')

        max_struct_speed_kts = max_struct_speed_mph / 1.15

        if CATD < 3:
            wing_loading = inputs[Aircraft.Wing.LOADING]

            VCMAX = 0.9 * max_struct_speed_kts
            VCCOF = 33.0
            if WGS_greater_than_20_flag:
                VCCOF = 33.0 - (0.0550 * (wing_loading - 20.0))

            if CATD == 2:
                VCCOF = 36.0
                if WGS_greater_than_20_flag:
                    VCCOF = 36.0 - (0.0925 * (wing_loading - 20.0))

            VCMIN = VCCOF * (wing_loading**0.5)

            if smooth:
                VCMIN = VCMIN * sig((VCMAX - VCMIN) / VCMAX) + VCMAX * sig(
                    (VCMIN - VCMAX) / VCMAX
                )

            else:
                if VCMIN > VCMAX:
                    VCMIN = VCMAX

            VDCOF = 1.4
            if WGS_greater_than_20_flag:
                VDCOF = 1.4 - (0.000625 * (wing_loading - 20.0))

            if CATD != 0:
                VDCOF = 1.5
                if WGS_greater_than_20_flag:
                    VDCOF = 1.5 - (0.001875 * (wing_loading - 20.0))

                if CATD != 1:
                    VDCOF = 1.55
                    if WGS_greater_than_20_flag:
                        VDCOF = 1.55 - (0.0025 * (wing_loading - 20.0))

            min_dive_vel = VDCOF * VCMIN

            if smooth:
                min_dive_vel = max_struct_speed_kts * sig(
                    (max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts
                ) + min_dive_vel * sig(
                    (min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts
                )
            else:
                if min_dive_vel < max_struct_speed_kts:
                    min_dive_vel = max_struct_speed_kts

            max_airspeed = 0.85 * min_dive_vel
            vel_c = VCMIN
            max_maneuver_factor = 3.8
            if CATD == 1.0:
                max_maneuver_factor = 4.4
            elif CATD == 2.0:
                max_maneuver_factor = 6.0

        elif CATD == 3:
            max_maneuver_factor = 2.5
            min_dive_vel = 1.2 * max_struct_speed_kts
            max_airspeed = max_struct_speed_kts
            vel_c = max_airspeed

        elif CATD > 3.001:
            max_maneuver_factor = CATD
            min_dive_vel = 1.2 * max_struct_speed_kts
            max_airspeed = max_struct_speed_kts
            vel_c = max_airspeed

        outputs["max_airspeed"] = max_airspeed
        outputs["vel_c"] = vel_c
        outputs["max_maneuver_factor"] = max_maneuver_factor
        outputs["min_dive_vel"] = min_dive_vel

    def compute_partials(self, inputs, partials):

        max_struct_speed_mph = inputs[Aircraft.Design.MAX_STRUCTURAL_SPEED]

        CATD = self.options["aviary_options"].get_val(
            Aircraft.Design.PART25_STRUCTURAL_CATEGORY, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')
        WGS_greater_than_20_flag = self.options["aviary_options"].get_val(
            Aircraft.Wing.LOADING_ABOVE_20, units='unitless')

        max_struct_speed_kts = max_struct_speed_mph / 1.15
        dmax_struct_speed_kts_dmax_struct_speed_mph = 1 / 1.15
        dmax_struct_speed_kts_dwing_loading = 0.0

        if CATD < 3:
            wing_loading = inputs[Aircraft.Wing.LOADING]
            VCMAX = 0.9 * max_struct_speed_kts
            VCCOF = 33.0
            dVCCOF_dwing_loading = 0.0
            dVCMAX_dmax_struct_speed_mph = 0.9 / 1.15

            if WGS_greater_than_20_flag:
                VCCOF = 33.0 - (0.0550 * (wing_loading - 20.0))
                dVCCOF_dwing_loading = -0.0550

            if CATD == 2:
                VCCOF = 36
                dVCCOF_dwing_loading = 0.0
                if WGS_greater_than_20_flag:
                    VCCOF = 36.0 - (0.0925 * (wing_loading - 20.0))
                    dVCCOF_dwing_loading = -0.0925

            VCMIN = VCCOF * (wing_loading**0.5)
            dVCMIN_dwing_loading = (
                dVCCOF_dwing_loading * (wing_loading**0.5)
                + VCCOF * 0.5 * wing_loading**-0.5
            )
            dVCMIN_dmax_struct_speed_mph = 0.0

            if smooth:
                VCMIN_1 = VCMIN * sig((VCMAX - VCMIN) / VCMAX) + VCMAX * sig(
                    (VCMIN - VCMAX) / VCMAX
                )
                dVCMIN_dwing_loading = (
                    dVCMIN_dwing_loading * sig((VCMAX - VCMIN) / VCMAX)
                    + VCMIN
                    * dsig((VCMAX - VCMIN) / VCMAX)
                    * -dVCMIN_dwing_loading
                    / VCMAX
                    + VCMAX
                    * dsig((VCMIN - VCMAX) / VCMAX)
                    * dVCMIN_dwing_loading
                    / VCMAX
                )
                dVCMIN_dmax_struct_speed_mph = (
                    dVCMIN_dmax_struct_speed_mph * sig((VCMAX - VCMIN) / VCMAX)
                    + VCMIN
                    * dsig((VCMAX - VCMIN) / VCMAX)
                    * dquotient(
                        (VCMAX - VCMIN),
                        VCMAX,
                        dVCMAX_dmax_struct_speed_mph - dVCMIN_dmax_struct_speed_mph,
                        dVCMAX_dmax_struct_speed_mph,
                    )
                    + dVCMAX_dmax_struct_speed_mph * sig((VCMIN - VCMAX) / VCMAX)
                    + VCMAX
                    * dsig((VCMIN - VCMAX) / VCMAX)
                    * dquotient(
                        (VCMIN - VCMAX),
                        VCMAX,
                        dVCMIN_dmax_struct_speed_mph - dVCMAX_dmax_struct_speed_mph,
                        dVCMAX_dmax_struct_speed_mph,
                    )
                )
                VCMIN = VCMIN_1
            else:
                if VCMIN > VCMAX:
                    VCMIN = VCMAX
                    dVCMIN_dmax_struct_speed_mph = dVCMAX_dmax_struct_speed_mph
                    dVCMIN_dwing_loading = 0.0

            if CATD == 1:

                dVCMIN_dwing_loading = 0.0
                dVCMIN_dmax_struct_speed_mph = 0.9 / 1.15

            VDCOF = 1.4
            dVDCOF_dwing_loading = 0.0

            if WGS_greater_than_20_flag:
                VDCOF = 1.4 - (0.000625 * (wing_loading - 20.0))
                dVDCOF_dwing_loading = -0.000625

            if CATD != 0:
                VDCOF = 1.5
                dVDCOF_dwing_loading = 0.0
                if WGS_greater_than_20_flag:
                    VDCOF = 1.5 - (0.001875 * (wing_loading - 20.0))
                    dVDCOF_dwing_loading = -0.001875

                if CATD != 1:
                    dVDCOF_dwing_loading = 0.0
                    if WGS_greater_than_20_flag:
                        VDCOF = 1.55 - (0.0025 * (wing_loading - 20.0))
                        dVDCOF_dwing_loading = -0.0025

            min_dive_vel = VDCOF * VCMIN
            dmin_dive_vel_dwing_loading = (
                dVDCOF_dwing_loading * VCMIN + VDCOF * dVCMIN_dwing_loading
            )
            dmin_dive_vel_dmax_struct_speed_mph = VDCOF * dVCMIN_dmax_struct_speed_mph

            if smooth:
                min_dive_vel_1 = max_struct_speed_kts * sig(
                    (max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts
                ) + min_dive_vel * sig(
                    (min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts
                )
                dmin_dive_vel_dmax_struct_speed_mph = (
                    dmax_struct_speed_kts_dmax_struct_speed_mph
                    * sig((max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts)
                    + max_struct_speed_kts
                    * dsig((max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts)
                    * dquotient(
                        (max_struct_speed_kts - min_dive_vel),
                        max_struct_speed_kts,
                        (
                            dmax_struct_speed_kts_dmax_struct_speed_mph
                            - dmin_dive_vel_dmax_struct_speed_mph
                        ),
                        dmax_struct_speed_kts_dmax_struct_speed_mph,
                    )
                    + dmin_dive_vel_dmax_struct_speed_mph
                    * sig((min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts)
                    + min_dive_vel
                    * dsig((min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts)
                    * dquotient(
                        (min_dive_vel - max_struct_speed_kts),
                        max_struct_speed_kts,
                        dmin_dive_vel_dmax_struct_speed_mph
                        - dmax_struct_speed_kts_dmax_struct_speed_mph,
                        dmax_struct_speed_kts_dmax_struct_speed_mph,
                    )
                )
                dmin_dive_vel_dwing_loading = (
                    dmax_struct_speed_kts_dwing_loading
                    * sig((max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts)
                    + max_struct_speed_kts
                    * dsig((max_struct_speed_kts - min_dive_vel) / max_struct_speed_kts)
                    * dquotient(
                        max_struct_speed_kts - min_dive_vel,
                        max_struct_speed_kts,
                        dmax_struct_speed_kts_dwing_loading
                        - dmin_dive_vel_dwing_loading,
                        dmax_struct_speed_kts_dwing_loading,
                    )
                    + dmin_dive_vel_dwing_loading
                    * sig((min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts)
                    + min_dive_vel
                    * dsig((min_dive_vel - max_struct_speed_kts) / max_struct_speed_kts)
                    * dquotient(
                        (min_dive_vel - max_struct_speed_kts),
                        max_struct_speed_kts,
                        dmin_dive_vel_dwing_loading
                        - dmax_struct_speed_kts_dwing_loading,
                        dmax_struct_speed_kts_dwing_loading,
                    )
                )
                min_dive_vel = min_dive_vel_1

            else:
                if (
                    min_dive_vel < max_struct_speed_kts
                ):  # note: this creates a discontinuity
                    min_dive_vel = max_struct_speed_kts
                    dmin_dive_vel_dwing_loading = 0
                    dmin_dive_vel_dmax_struct_speed_mph = (
                        dmax_struct_speed_kts_dmax_struct_speed_mph
                    )

            partials[
                "min_dive_vel", Aircraft.Wing.LOADING
            ] = dmin_dive_vel_dwing_loading
            partials[
                "min_dive_vel", Aircraft.Design.MAX_STRUCTURAL_SPEED
            ] = dmin_dive_vel_dmax_struct_speed_mph

            partials["max_airspeed", Aircraft.Wing.LOADING] = (
                0.85 * dmin_dive_vel_dwing_loading
            )
            partials["max_airspeed", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                0.85 * dmin_dive_vel_dmax_struct_speed_mph
            )

            partials["vel_c", Aircraft.Wing.LOADING] = dVCMIN_dwing_loading
            partials["vel_c", Aircraft.Design.MAX_STRUCTURAL_SPEED] = dVCMIN_dmax_struct_speed_mph

        if CATD == 3:

            partials[
                "max_airspeed", Aircraft.Design.MAX_STRUCTURAL_SPEED
            ] = dmax_struct_speed_kts_dmax_struct_speed_mph
            partials["min_dive_vel", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                1.2 * dmax_struct_speed_kts_dmax_struct_speed_mph
            )
            partials["vel_c", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                1.0 * dmax_struct_speed_kts_dmax_struct_speed_mph
            )

        elif CATD > 3.001:

            partials["max_airspeed", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                1 * dmax_struct_speed_kts_dmax_struct_speed_mph
            )
            partials["min_dive_vel", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                1.2 * dmax_struct_speed_kts_dmax_struct_speed_mph
            )
            partials["vel_c", Aircraft.Design.MAX_STRUCTURAL_SPEED] = (
                1.0 * dmax_struct_speed_kts_dmax_struct_speed_mph
            )


class LoadParameters(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        self.add_input(
            "vel_c",
            val=100,
            units="kn",
            desc="VGC: Velocity used in Gust Load Factor calculation at cruise conditions.\
                This is Minimum Design Cruise Speed for Part 23 aircraft and VM0 for Part 25 aircraft",
        )
        self.add_input(
            "max_airspeed",
            val=200,
            units="kn",
            desc="VM0: maximum operating equivalent airspeed",
        )

        self.add_output(
            "max_mach", val=0, units="unitless", desc="EMM0: maximum operating mach number"
        )
        self.add_output(
            "density_ratio",
            val=0,
            units="unitless",
            desc="SIGMA (in GASP): density ratio = density at Altitude / density at Sea level",
        )
        self.add_output(
            "V9", val=0, units="kn", desc="V9: intermediate value. Typically it is maximum flight speed."
        )

        self.declare_partials("max_mach", "max_airspeed", val=0.0)
        self.declare_partials("density_ratio", "max_airspeed", val=0.0)
        self.declare_partials("V9", "*", val=0.0)

    def compute(self, inputs, outputs):

        vel_c = inputs["vel_c"]
        max_airspeed = inputs["max_airspeed"]

        cruise_alt = self.options["aviary_options"].get_val(
            Mission.Design.CRUISE_ALTITUDE, units='ft')
        CATD = self.options["aviary_options"].get_val(
            Aircraft.Design.PART25_STRUCTURAL_CATEGORY, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        if cruise_alt <= 22500.0:
            max_mach = max_airspeed / 486.33
        if cruise_alt > 22500.0 and cruise_alt <= 36000.0:
            max_mach = max_airspeed / 424.73
            dmax_mach_dmax_airspeed = 1 / 424.73
        if cruise_alt > 36000.0:
            max_mach = max_airspeed / 372.34

        if smooth:
            max_mach = max_mach * sig((0.9 - max_mach) / 0.9) + 0.9 * sig(
                (max_mach - 0.9) / 0.9
            )

        else:
            if max_mach > 0.90:  # note: this creates a discontinuity
                max_mach = 0.90

        density_ratio = (max_airspeed / (661.7 * max_mach)) ** 1.61949

        if smooth:
            V9 = vel_c * sig(1 - density_ratio) + 661.7 * \
                max_mach * sig(density_ratio - 1)

            if CATD < 3:
                # this line creates a smooth bounded density_ratio such that .6820<=density_ratio<=1
                density_ratio = (
                    0.6820 * sig((0.6820 - density_ratio) / 0.6820)
                    + density_ratio * sig((density_ratio - 0.6820) /
                                          0.6820) * sig(1 - density_ratio)
                    + sig(density_ratio - 1)
                )

            else:
                # this line creates a smooth bounded density_ratio such that .53281<=density_ratio<=1
                density_ratio = (
                    0.53281 * sig((0.53281 - density_ratio) / 0.53281)
                    + density_ratio * sig((density_ratio - 0.53281) /
                                          0.53281) * sig(1 - density_ratio)
                    + sig(density_ratio - 1)
                )

        else:
            if density_ratio >= 0.53281:  # note: this creates a discontinuity
                V9 = vel_c
                if density_ratio > 1:  # note: this creates a discontinuity
                    V9 = 661.7 * max_mach
                    density_ratio = 1.0

            else:  # note: this creates a discontinuity
                density_ratio = 0.53281
                V9 = vel_c

            if CATD < 3.0 and density_ratio <= 0.6820:  # note: this creates a discontinuity
                density_ratio = 0.6820

        outputs["max_mach"] = max_mach
        outputs["density_ratio"] = density_ratio
        outputs["V9"] = V9

    def compute_partials(self, inputs, partials):

        vel_c = inputs["vel_c"]
        max_airspeed = inputs["max_airspeed"]

        cruise_alt = self.options["aviary_options"].get_val(
            Mission.Design.CRUISE_ALTITUDE, units='ft')
        CATD = self.options["aviary_options"].get_val(
            Aircraft.Design.PART25_STRUCTURAL_CATEGORY, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        if cruise_alt <= 22500.0:
            max_mach = max_airspeed / 486.33
            dmax_mach_dmax_airspeed = 1 / 486.33
        if cruise_alt > 22500.0 and cruise_alt <= 36000.0:
            max_mach = max_airspeed / 424.73
            dmax_mach_dmax_airspeed = 1 / 424.73
        if cruise_alt > 36000.0:
            max_mach = max_airspeed / 372.34
            dmax_mach_dmax_airspeed = 1 / 372.34

        if smooth:
            max_mach_1 = max_mach * sig((0.9 - max_mach) / 0.9) + 0.9 * sig(
                (max_mach - 0.9) / 0.9
            )
            dmax_mach_dmax_airspeed = (
                dmax_mach_dmax_airspeed * sig((0.9 - max_mach) / 0.9)
                + max_mach
                * dsig((0.9 - max_mach) / 0.9)
                * dquotient((0.9 - max_mach), 0.9, -dmax_mach_dmax_airspeed, 0.0)
                + 0.9
                * dsig((max_mach - 0.9) / 0.9)
                * dquotient((max_mach - 0.9), 0.9, dmax_mach_dmax_airspeed, 0.0)
            )
            max_mach = max_mach_1
        else:
            if max_mach > 0.90:  # note: this creates a discontinuity
                max_mach = 0.90
                dmax_mach_dmax_airspeed = 0.0

        density_ratio = (max_airspeed / (661.7 * max_mach)) ** 1.61949
        ddensity_ratio_dmax_airspeed = (
            1.61949
            * (max_airspeed / (661.7 * max_mach)) ** 0.61949
            * dquotient(
                max_airspeed, (661.7 * max_mach), 1, 661.7 * dmax_mach_dmax_airspeed
            )
        )

        if smooth:
            V9_1 = vel_c * sig(1 - density_ratio) + 661.7 * \
                max_mach * sig(density_ratio - 1)
            dV9_dmax_airspeed = (
                vel_c * dsig(1 - density_ratio) * -ddensity_ratio_dmax_airspeed
                + 661.7
                * dmax_mach_dmax_airspeed
                * sig(density_ratio - 1)
                * 661.7
                * max_mach
                * dsig(density_ratio - 1)
                * ddensity_ratio_dmax_airspeed
            )
            dV9_dvel_c = sig(1 - density_ratio)
            V9 = V9_1

            if CATD < 3:
                # this line creates a smooth bounded density_ratio such that .6820<=density_ratio<=1
                density_ratio_1 = (
                    0.6820 * sig((0.6820 - density_ratio) / 0.6820)
                    + density_ratio * sig((density_ratio - 0.6820) /
                                          0.6820) * sig(1 - density_ratio)
                    + sig(density_ratio - 1)
                )
                ddensity_ratio_dmax_airspeed = (
                    0.6820
                    * dsig((0.6820 - density_ratio) / 0.6820)
                    * -ddensity_ratio_dmax_airspeed
                    / 0.6820
                    + ddensity_ratio_dmax_airspeed
                    * sig((density_ratio - 0.6820) / 0.6820)
                    * sig(1 - density_ratio)
                    + density_ratio
                    * (
                        dsig((density_ratio - 0.6820) / 0.6820)
                        * ddensity_ratio_dmax_airspeed
                        / 0.6820
                        * sig(1 - density_ratio)
                        + sig((density_ratio - 0.6820) / 0.6820)
                        * dsig(1 - density_ratio)
                        * -ddensity_ratio_dmax_airspeed
                    )
                    + dsig(density_ratio - 1) * ddensity_ratio_dmax_airspeed
                )
                density_ratio = density_ratio_1

            else:
                # this line creates a smooth bounded density_ratio such that .53281<=density_ratio<=1
                density_ratio_1 = (
                    0.53281 * sig((0.53281 - density_ratio) / 0.53281)
                    + density_ratio * sig((density_ratio - 0.53281) /
                                          0.53281) * sig(1 - density_ratio)
                    + sig(density_ratio - 1)
                )
                ddensity_ratio_dmax_airspeed = (
                    0.53281
                    * dsig((0.53281 - density_ratio) / 0.53281)
                    * -ddensity_ratio_dmax_airspeed
                    / 0.53281
                    + ddensity_ratio_dmax_airspeed
                    * sig((density_ratio - 0.53281) / 0.53281)
                    * sig(1 - density_ratio)
                    + density_ratio
                    * (
                        dsig((density_ratio - 0.53281) / 0.53281)
                        * ddensity_ratio_dmax_airspeed
                        / 0.53281
                        * sig(1 - density_ratio)
                        + sig((density_ratio - 0.53281) / 0.53281)
                        * dsig(1 - density_ratio)
                        * -ddensity_ratio_dmax_airspeed
                    )
                    + dsig(density_ratio - 1) * ddensity_ratio_dmax_airspeed
                )
                density_ratio = density_ratio_1
        else:
            if density_ratio >= 0.53281:  # note: this creates a discontinuity
                V9 = vel_c
                dV9_dvel_c = 1.0
                dV9_dmax_airspeed = 0.0

                if density_ratio > 1:  # note: this creates a discontinuity
                    V9 = 661.7 * max_mach
                    density_ratio = 1.0

                    dV9_dvel_c = 0.0
                    dV9_dmax_airspeed = 661.7 * dmax_mach_dmax_airspeed
                    ddensity_ratio_dmax_airspeed = 0.0

            else:  # note: this creates a discontinuity
                density_ratio = 0.53281
                V9 = vel_c

                dV9_dvel_c = 1.0
                dV9_dmax_airspeed = 0.0
                ddensity_ratio_dmax_airspeed = 0.0

            if CATD < 3.0 and density_ratio <= 0.6820:  # note: this creates a discontinuity
                density_ratio = 0.6820
                ddensity_ratio_dmax_airspeed = 0.0

        partials["max_mach", "max_airspeed"] = dmax_mach_dmax_airspeed
        partials["density_ratio", "max_airspeed"] = ddensity_ratio_dmax_airspeed
        partials["V9", "max_airspeed"] = dV9_dmax_airspeed
        partials["V9", "vel_c"] = dV9_dvel_c


class LiftCurveSlopeAtCruise(om.ExplicitComponent):
    def setup(self):
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)
        add_aviary_input(self, Aircraft.Wing.SWEEP, val=0.436, units="rad")
        add_aviary_input(self, Mission.Design.MACH, val=0.8)

        add_aviary_output(self, Aircraft.Design.LIFT_CURVE_SLOPE, val=7.1765)

        self.declare_partials(Aircraft.Design.LIFT_CURVE_SLOPE, "*")

    def compute(self, inputs, outputs):
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        DLMC4 = inputs[Aircraft.Wing.SWEEP]
        mach = inputs[Mission.Design.MACH]

        outputs[Aircraft.Design.LIFT_CURVE_SLOPE] = (
            np.pi * AR / (1 + np.sqrt(1 + (((AR / (2 * np.cos(DLMC4))) ** 2) * (1 - (mach * np.cos(DLMC4)) ** 2)))))

    def compute_partials(self, inputs, partials):
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        DLMC4 = inputs[Aircraft.Wing.SWEEP]
        mach = inputs[Mission.Design.MACH]

        c1 = np.sqrt(AR**2 * (-mach**2*np.cos(DLMC4)**2 + 1) + 4*np.cos(DLMC4)**2)
        c2 = 2*np.cos(DLMC4) + c1

        partials[Aircraft.Design.LIFT_CURVE_SLOPE, Aircraft.Wing.ASPECT_RATIO] = (
            4 * np.pi * np.cos(DLMC4)**2) / (c1 * c2)
        partials[Aircraft.Design.LIFT_CURVE_SLOPE, Mission.Design.MACH] = (
            2 * np.pi * AR**3 * mach * np.cos(DLMC4)**3) / (c1 * c2**2)
        partials[Aircraft.Design.LIFT_CURVE_SLOPE, Aircraft.Wing.SWEEP] = (-np.pi * AR * (-8 * np.cos(DLMC4)**3 * np.sin(
            DLMC4) + 4 * np.cos(DLMC4)**2 * np.sin(2*DLMC4) + AR**2 * np.sin(2*DLMC4))) / (np.cos(DLMC4) * c1 * c2**2)


class LoadFactors(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Wing.LOADING, val=128)

        self.add_input(
            "density_ratio",
            val=0.5,
            units="unitless",
            desc="SIGMA (in GASP): density ratio = density at Altitude / density at Sea level",
        )
        self.add_input(
            "V9",
            val=100,
            units="kn",
            desc="V9: intermediate value. Typically it is maximum flight speed.",
        )
        self.add_input("min_dive_vel", val=250, units="kn", desc="VDMIN: dive velocity")
        self.add_input(
            "max_maneuver_factor",
            val=0.72,
            units="unitless",
            desc="EMLF: maximum maneuver load factor, units are in g`s",
        )

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=12.6131)
        add_aviary_input(self, Aircraft.Design.LIFT_CURVE_SLOPE, val=7.1765)

        add_aviary_output(self, Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.5)

        self.declare_partials(Aircraft.Wing.ULTIMATE_LOAD_FACTOR, "*", val=0.0)

    def compute(self, inputs, outputs):

        wing_loading = inputs[Aircraft.Wing.LOADING]
        density_ratio = inputs["density_ratio"]
        V9 = inputs["V9"]
        min_dive_vel = inputs["min_dive_vel"]
        max_maneuver_factor = inputs["max_maneuver_factor"]
        avg_chord = inputs[Aircraft.Wing.AVERAGE_CHORD]
        Cl_alpha = inputs[Aircraft.Design.LIFT_CURVE_SLOPE]

        ULF_from_maneuver = self.options["aviary_options"].get_val(
            Aircraft.Design.ULF_CALCULATED_FROM_MANEUVER, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        mass_ratio = (
            2.0 * wing_loading / (density_ratio * 0.00237 * avg_chord * Cl_alpha * 32.2)
        )
        k_load_factor = 0.88 * mass_ratio / (5.3 + mass_ratio)
        cruise_load_factor = 1.0 + (
            (k_load_factor * 50.0 * V9 * Cl_alpha) / (498.0 * wing_loading)
        )
        dive_load_factor = 1.0 + (
            (k_load_factor * 25.0 * min_dive_vel * Cl_alpha) / (498.0 * wing_loading)
        )
        gust_load_factor = dive_load_factor

        if smooth:
            gust_load_factor = dive_load_factor * sig(
                (dive_load_factor - cruise_load_factor) / dive_load_factor
            ) + cruise_load_factor * sig(
                (cruise_load_factor - dive_load_factor) / dive_load_factor
            )

        else:
            if (
                cruise_load_factor > dive_load_factor
            ):  # note: this creates a discontinuity
                gust_load_factor = cruise_load_factor

        ULF = 1.5 * max_maneuver_factor

        if smooth:
            ULF = 1.5 * (
                gust_load_factor
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + max_maneuver_factor
                * sig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
            )

        else:
            if (
                gust_load_factor > max_maneuver_factor
            ):  # note: this creates a discontinuity
                ULF = 1.5 * gust_load_factor

        if ULF_from_maneuver == True:
            ULF = 1.5 * max_maneuver_factor

        outputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR] = ULF

    def compute_partials(self, inputs, partials):

        wing_loading = inputs[Aircraft.Wing.LOADING]
        density_ratio = inputs["density_ratio"]
        V9 = inputs["V9"]
        min_dive_vel = inputs["min_dive_vel"]
        max_maneuver_factor = inputs["max_maneuver_factor"]
        avg_chord = inputs[Aircraft.Wing.AVERAGE_CHORD]
        Cl_alpha = inputs[Aircraft.Design.LIFT_CURVE_SLOPE]

        ULF_from_maneuver = self.options["aviary_options"].get_val(
            Aircraft.Design.ULF_CALCULATED_FROM_MANEUVER, units='unitless')
        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        mass_ratio = (
            2.0 * wing_loading / (density_ratio * 0.00237 * avg_chord * Cl_alpha * 32.2)
        )
        k_load_factor = 0.88 * mass_ratio / (5.3 + mass_ratio)
        cruise_load_factor = 1.0 + (
            (k_load_factor * 50.0 * V9 * Cl_alpha) / (498.0 * wing_loading)
        )
        dive_load_factor = 1.0 + (
            (k_load_factor * 25.0 * min_dive_vel * Cl_alpha) / (498.0 * wing_loading)
        )
        gust_load_factor = dive_load_factor

        dmass_ratio_dwing_loading = 2.0 / (
            density_ratio * 0.00237 * avg_chord * Cl_alpha * 32.2
        )
        dmass_ratio_ddensity_ratio = (
            -2.0 * wing_loading / (density_ratio**2 * 0.00237 *
                                   avg_chord * Cl_alpha * 32.2)
        )
        dmass_ratio_davg_chord = (
            -2.0 * wing_loading / (density_ratio * 0.00237 *
                                   avg_chord**2 * Cl_alpha * 32.2)
        )
        dmass_ratio_dCl_alpha = (
            -2.0 * wing_loading / (density_ratio * 0.00237 *
                                   avg_chord * Cl_alpha**2 * 32.2)
        )

        dk_load_factor_dwing_loading = dquotient(
            0.88 * mass_ratio,
            5.3 + mass_ratio,
            0.88 * dmass_ratio_dwing_loading,
            dmass_ratio_dwing_loading,
        )
        dk_load_factor_ddensity_ratio = dquotient(
            0.88 * mass_ratio,
            5.3 + mass_ratio,
            0.88 * dmass_ratio_ddensity_ratio,
            dmass_ratio_ddensity_ratio,
        )
        dk_load_factor_davg_chord = dquotient(
            0.88 * mass_ratio,
            5.3 + mass_ratio,
            0.88 * dmass_ratio_davg_chord,
            dmass_ratio_davg_chord,
        )
        dk_load_factor_dCl_alpha = dquotient(
            0.88 * mass_ratio,
            5.3 + mass_ratio,
            0.88 * dmass_ratio_dCl_alpha,
            dmass_ratio_dCl_alpha,
        )

        dcruise_load_factor_dwing_loading = dquotient(
            (k_load_factor * 50.0 * V9 * Cl_alpha),
            (498.0 * wing_loading),
            (dk_load_factor_dwing_loading * 50.0 * V9 * Cl_alpha),
            498,
        )
        dcruise_load_factor_ddensity_ratio = (dk_load_factor_ddensity_ratio * 50.0 * V9 * Cl_alpha) / (
            498.0 * wing_loading
        )
        dcruise_load_factor_davg_chord = (
            dk_load_factor_davg_chord * 50.0 * V9 * Cl_alpha
        ) / (498.0 * wing_loading)
        dcruise_load_factor_dCl_alpha = (
            dk_load_factor_dCl_alpha * 50.0 * V9 * Cl_alpha + k_load_factor * 50 * V9
        ) / (498.0 * wing_loading)
        dcruise_load_factor_dV9 = (k_load_factor * 50.0 * Cl_alpha) / (
            498.0 * wing_loading
        )
        dcruise_load_factor_dmin_dive_vel = 0.0

        ddive_load_factor_dwing_loading = dquotient(
            (k_load_factor * 25.0 * min_dive_vel * Cl_alpha),
            (498.0 * wing_loading),
            (dk_load_factor_dwing_loading * 25.0 * min_dive_vel * Cl_alpha),
            498,
        )
        ddive_load_factor_ddensity_ratio = (
            dk_load_factor_ddensity_ratio * 25 * min_dive_vel * Cl_alpha
        ) / (498.0 * wing_loading)
        ddive_load_factor_davg_chord = (
            dk_load_factor_davg_chord * 25 * min_dive_vel * Cl_alpha
        ) / (498.0 * wing_loading)
        ddive_load_factor_dCl_alpha = (
            dk_load_factor_dCl_alpha * 25 * min_dive_vel * Cl_alpha
            + k_load_factor * 50 * min_dive_vel
        ) / (498.0 * wing_loading)
        ddive_load_factor_dmin_dive_vel = (k_load_factor * 25 * Cl_alpha) / (
            498.0 * wing_loading
        )
        ddive_load_factor_dV9 = 0.0

        dgust_load_factor_dwing_loading = dquotient(
            (k_load_factor * 25.0 * min_dive_vel * Cl_alpha),
            (498.0 * wing_loading),
            (dk_load_factor_dwing_loading * 25.0 * min_dive_vel * Cl_alpha),
            498,
        )
        dgust_load_factor_ddensity_ratio = (
            dk_load_factor_ddensity_ratio * 25 * min_dive_vel * Cl_alpha
        ) / (498.0 * wing_loading)
        dgust_load_factor_davg_chord = (
            dk_load_factor_davg_chord * 25 * min_dive_vel * Cl_alpha
        ) / (498.0 * wing_loading)
        dgust_load_factor_dCl_alpha = (
            dk_load_factor_dCl_alpha * 25 * min_dive_vel * Cl_alpha
            + k_load_factor * 50 * min_dive_vel
        ) / (498.0 * wing_loading)
        dgust_load_factor_dmin_dive_vel = (k_load_factor * 25 * Cl_alpha) / (
            498.0 * wing_loading
        )
        dgust_load_factor_dV9 = 0.0

        if smooth:
            gust_load_factor_1 = dive_load_factor * sig(
                (dive_load_factor - cruise_load_factor) / dive_load_factor
            ) + cruise_load_factor * sig(
                (cruise_load_factor - dive_load_factor) / dive_load_factor
            )
            dgust_load_factor_dwing_loading = (
                ddive_load_factor_dwing_loading
                * sig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                + dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_dwing_loading - dcruise_load_factor_dwing_loading,
                    ddive_load_factor_dwing_loading,
                )
                + dcruise_load_factor_dwing_loading
                * sig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    dcruise_load_factor_dwing_loading - ddive_load_factor_dwing_loading,
                    ddive_load_factor_dwing_loading,
                )
            )
            dgust_load_factor_ddensity_ratio = (
                ddive_load_factor_ddensity_ratio
                * sig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                + dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_ddensity_ratio - dcruise_load_factor_ddensity_ratio,
                    ddive_load_factor_ddensity_ratio,
                )
                + dcruise_load_factor_ddensity_ratio
                * sig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    dcruise_load_factor_ddensity_ratio - ddive_load_factor_ddensity_ratio,
                    ddive_load_factor_ddensity_ratio,
                )
            )
            dgust_load_factor_davg_chord = (
                ddive_load_factor_davg_chord
                * sig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                + dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_davg_chord - dcruise_load_factor_davg_chord,
                    ddive_load_factor_davg_chord,
                )
                + dcruise_load_factor_davg_chord
                * sig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    dcruise_load_factor_davg_chord - ddive_load_factor_davg_chord,
                    ddive_load_factor_davg_chord,
                )
            )
            dgust_load_factor_dCl_alpha = (
                ddive_load_factor_dCl_alpha
                * sig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                + dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_dCl_alpha - dcruise_load_factor_dCl_alpha,
                    ddive_load_factor_dCl_alpha,
                )
                + dcruise_load_factor_dCl_alpha
                * sig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    dcruise_load_factor_dCl_alpha - ddive_load_factor_dCl_alpha,
                    ddive_load_factor_dCl_alpha,
                )
            )
            dgust_load_factor_dV9 = (
                dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_dV9 - dcruise_load_factor_dV9,
                    ddive_load_factor_dV9,
                )
                + dcruise_load_factor_dV9
                * sig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    dcruise_load_factor_dV9 - ddive_load_factor_dV9,
                    ddive_load_factor_dV9,
                )
            )
            dgust_loading_dmin_dive_vel = (
                ddive_load_factor_dmin_dive_vel
                * sig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                + dive_load_factor
                * dsig((dive_load_factor - cruise_load_factor) / dive_load_factor)
                * dquotient(
                    (dive_load_factor - cruise_load_factor),
                    dive_load_factor,
                    ddive_load_factor_dmin_dive_vel,
                    ddive_load_factor_dmin_dive_vel,
                )
                + cruise_load_factor
                * dsig((cruise_load_factor - dive_load_factor) / dive_load_factor)
                * dquotient(
                    (cruise_load_factor - dive_load_factor),
                    dive_load_factor,
                    -ddive_load_factor_dmin_dive_vel,
                    ddive_load_factor_dmin_dive_vel,
                )
            )
            gust_load_factor = gust_load_factor_1
        else:

            if (
                cruise_load_factor > dive_load_factor
            ):  # note: this creates a discontinuity
                gust_load_factor = cruise_load_factor

                dgust_load_factor_dwing_loading = dcruise_load_factor_dwing_loading
                dgust_load_factor_ddensity_ratio = dcruise_load_factor_ddensity_ratio
                dgust_load_factor_davg_chord = dcruise_load_factor_davg_chord
                dgust_load_factor_dCl_alpha = dcruise_load_factor_dCl_alpha
                dgust_load_factor_dV9 = dcruise_load_factor_dV9
                dgust_load_factor_dmin_dive_vel = 0.0

        ULF = 1.5 * max_maneuver_factor
        dULF_dmax_maneuver_factor = 1.5
        dULF_dwing_loading = 0.0
        dULF_ddensity_ratio = 0.0
        dULF_davg_chord = 0.0
        dULF_dCl_alpha = 0.0
        dULF_dV9 = 0.0
        dULF_dmin_dive_vel = 0.0

        if smooth:
            ULF_1 = 1.5 * (
                gust_load_factor
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + max_maneuver_factor
                * sig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
            )
            dULF_dmax_maneuver_factor = 1.5 * (
                gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    -1.0,
                    0.0,
                )
                + sig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor), gust_load_factor, 1.0, 0.0
                )
            )
            dULF_dwing_loading = 1.5 * (
                dgust_load_factor_dwing_loading
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    dgust_load_factor_dwing_loading,
                    dgust_load_factor_dwing_loading,
                )
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor),
                    gust_load_factor,
                    -dgust_load_factor_dwing_loading,
                    dgust_load_factor_dwing_loading,
                )
            )
            dULF_ddensity_ratio = 1.5 * (
                dgust_load_factor_ddensity_ratio
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    dgust_load_factor_ddensity_ratio,
                    dgust_load_factor_ddensity_ratio,
                )
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor),
                    gust_load_factor,
                    -dgust_load_factor_ddensity_ratio,
                    dgust_load_factor_ddensity_ratio,
                )
            )
            dULF_davg_chord = 1.5 * (
                dgust_load_factor_davg_chord
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    dgust_load_factor_davg_chord,
                    dgust_load_factor_davg_chord,
                )
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor),
                    gust_load_factor,
                    -dgust_load_factor_davg_chord,
                    dgust_load_factor_davg_chord,
                )
            )
            dULF_dCl_alpha = 1.5 * (
                dgust_load_factor_dCl_alpha
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    dgust_load_factor_dCl_alpha,
                    dgust_load_factor_dCl_alpha,
                )
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor),
                    gust_load_factor,
                    -dgust_load_factor_dCl_alpha,
                    dgust_load_factor_dCl_alpha,
                )
            )
            dULF_dV9 = 1.5 * (
                dgust_load_factor_dV9
                * sig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                + gust_load_factor
                * dsig((gust_load_factor - max_maneuver_factor) / gust_load_factor)
                * dquotient(
                    (gust_load_factor - max_maneuver_factor),
                    gust_load_factor,
                    dgust_load_factor_dV9,
                    dgust_load_factor_dV9,
                )
                + max_maneuver_factor
                * dsig((max_maneuver_factor - gust_load_factor) / gust_load_factor)
                * dquotient(
                    (max_maneuver_factor - gust_load_factor),
                    gust_load_factor,
                    -dgust_load_factor_dV9,
                    dgust_load_factor_dV9,
                )
            )
            dULF_dmin_dive_vel = 0.0
            ULF = ULF_1
        else:
            if (
                gust_load_factor > max_maneuver_factor
            ):  # note: this creates a discontinuity
                ULF = 1.5 * gust_load_factor

                dULF_dmax_maneuver_factor = 0.0
                dULF_dwing_loading = 1.5 * dgust_load_factor_dwing_loading
                dULF_ddensity_ratio = 1.5 * dgust_load_factor_ddensity_ratio
                dULF_davg_chord = 1.5 * dgust_load_factor_davg_chord
                dULF_dCl_alpha = 1.5 * dgust_load_factor_dCl_alpha
                dULF_dV9 = 1.5 * dgust_load_factor_dV9
                dULF_dmin_dive_vel = 1.5 * dgust_load_factor_dmin_dive_vel

        if ULF_from_maneuver == True:
            ULF = 1.5 * max_maneuver_factor

            dULF_dmax_maneuver_factor = 1.5
            dULF_dwing_loading = 0.0
            dULF_ddensity_ratio = 0.0
            dULF_davg_chord = 0.0
            dULF_dCl_alpha = 0.0
            dULF_dV9 = 0.0
            dULF_dmin_dive_vel = 0.0

        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                 "max_maneuver_factor"] = dULF_dmax_maneuver_factor
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                 Aircraft.Wing.LOADING] = dULF_dwing_loading
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                 "density_ratio"] = dULF_ddensity_ratio
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                 Aircraft.Wing.AVERAGE_CHORD] = dULF_davg_chord
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                 Aircraft.Design.LIFT_CURVE_SLOPE] = dULF_dCl_alpha
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR, "V9"] = dULF_dV9
        partials[Aircraft.Wing.ULTIMATE_LOAD_FACTOR, "min_dive_vel"] = dULF_dmin_dive_vel


class DesignLoadGroup(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            "speeds",
            LoadSpeeds(
                aviary_options=aviary_options,
            ),
            promotes_inputs=['aircraft:*'],
            promotes_outputs=[
                "max_airspeed",
                "vel_c",
                "max_maneuver_factor",
                "min_dive_vel",
            ],
        )

        self.add_subsystem(
            "params",
            LoadParameters(
                aviary_options=aviary_options,
            ),
            promotes_inputs=["max_airspeed", "vel_c"],
            promotes_outputs=["density_ratio", "V9", "max_mach"],
        )

        self.add_subsystem(
            "Cl_Alpha_calc",
            LiftCurveSlopeAtCruise(),
            promotes_inputs=["aircraft:*", "mission:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.add_subsystem(
            "factors",
            LoadFactors(
                aviary_options=aviary_options,
            ),
            promotes_inputs=[
                "max_maneuver_factor",
                "min_dive_vel",
                "density_ratio",
                "V9",
            ]
            + ["aircraft:*"],
            promotes_outputs=[
                "aircraft:*"
            ],
        )
        self.set_input_defaults(Aircraft.Wing.LOADING, val=128, units="lbf/ft**2")



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, get_units
from aviary.variable_info.variables import Aircraft


class MuxComponent(om.ExplicitComponent):
    """
    Assemble component-wise vectors for wetted area, fineness, characteristic length,
    upper-surface laminar fraction, and lower-surface laminar fraction.

    Output ordering matches FLOPS:
    Wing, Horizontal Tail, Vertical Tails, Fuelsages, Nacelles
    """

    def __init__(self, **kwargs):
        self.num_tails = 0
        self.num_fuselages = 0
        self.num_nacelles = 0

        super().__init__(**kwargs)

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nc = 2
        aviary_options: AviaryValues = self.options['aviary_options']

        # Wing (Always 1)
        add_aviary_input(self, Aircraft.Wing.WETTED_AREA, 1.0)
        add_aviary_input(self, Aircraft.Wing.FINENESS, 1.0)
        add_aviary_input(self, Aircraft.Wing.CHARACTERISTIC_LENGTH, 1.0)
        add_aviary_input(self, Aircraft.Wing.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.Wing.LAMINAR_FLOW_LOWER, 0.0)

        # Horizontal Tail (Always 1)
        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA, 1.0)
        add_aviary_input(self, Aircraft.HorizontalTail.FINENESS, 1.0)
        add_aviary_input(self, Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH, 1.0)
        add_aviary_input(self, Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER, 0.0)

        zero_count = (0, None)
        # Vertical Tail
        num, _ = aviary_options.get_item(Aircraft.VerticalTail.NUM_TAILS, zero_count)
        self.num_tails = num
        if num > 0:
            add_aviary_input(self, Aircraft.VerticalTail.WETTED_AREA, 1.0)
            add_aviary_input(self, Aircraft.VerticalTail.FINENESS, 1.0)
            add_aviary_input(self, Aircraft.VerticalTail.CHARACTERISTIC_LENGTH, 1.0)
            add_aviary_input(self, Aircraft.VerticalTail.LAMINAR_FLOW_UPPER, 0.0)
            add_aviary_input(self, Aircraft.VerticalTail.LAMINAR_FLOW_LOWER, 0.0)
            nc += num

        # Fuselage
        num, _ = aviary_options.get_item(Aircraft.Fuselage.NUM_FUSELAGES, zero_count)
        self.num_fuselages = num
        if num > 0:
            add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, 1.0)
            add_aviary_input(self, Aircraft.Fuselage.FINENESS, 1.0)
            add_aviary_input(self, Aircraft.Fuselage.CHARACTERISTIC_LENGTH, 1.0)
            add_aviary_input(self, Aircraft.Fuselage.LAMINAR_FLOW_UPPER, 0.0)
            add_aviary_input(self, Aircraft.Fuselage.LAMINAR_FLOW_LOWER, 0.0)
            nc += num

        num, _ = aviary_options.get_item(
            Aircraft.Engine.NUM_ENGINES, zero_count)
        self.num_nacelles = int(sum(num))
        if any(num > 0):
            add_aviary_input(self, Aircraft.Nacelle.WETTED_AREA,
                             np.zeros(len(num)))
            add_aviary_input(self, Aircraft.Nacelle.FINENESS,
                             np.zeros(len(num)))
            add_aviary_input(self, Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
                             np.zeros(len(num)))
            add_aviary_input(self, Aircraft.Nacelle.LAMINAR_FLOW_UPPER,
                             np.zeros(len(num)))
            add_aviary_input(self, Aircraft.Nacelle.LAMINAR_FLOW_LOWER,
                             np.zeros(len(num)))
            nc += self.num_nacelles

        self.add_output(
            'wetted_areas', shape=nc,
            units=get_units(Aircraft.Wing.WETTED_AREA))
        self.add_output(
            'fineness_ratios', shape=nc,
            units=get_units(Aircraft.Wing.FINENESS))
        self.add_output(
            'characteristic_lengths', shape=nc,
            units=get_units(Aircraft.Wing.CHARACTERISTIC_LENGTH))
        self.add_output(
            'laminar_fractions_upper', shape=nc,
            units=get_units(Aircraft.Wing.LAMINAR_FLOW_UPPER))
        self.add_output(
            'laminar_fractions_lower', shape=nc,
            units=get_units(Aircraft.Wing.LAMINAR_FLOW_LOWER))

    def setup_partials(self):
        rows = np.zeros(1)
        cols = np.zeros(1)

        # Wing
        self.declare_partials(
            'wetted_areas', Aircraft.Wing.WETTED_AREA,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'fineness_ratios', Aircraft.Wing.FINENESS,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'characteristic_lengths', Aircraft.Wing.CHARACTERISTIC_LENGTH,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'laminar_fractions_upper', Aircraft.Wing.LAMINAR_FLOW_UPPER,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'laminar_fractions_lower', Aircraft.Wing.LAMINAR_FLOW_LOWER,
            rows=rows, cols=cols, val=1.0)

        # Horizontal Tail
        rows = np.ones(1)
        self.declare_partials(
            'wetted_areas', Aircraft.HorizontalTail.WETTED_AREA,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'fineness_ratios', Aircraft.HorizontalTail.FINENESS,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'characteristic_lengths', Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'laminar_fractions_upper', Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER,
            rows=rows, cols=cols, val=1.0)
        self.declare_partials(
            'laminar_fractions_lower', Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER,
            rows=rows, cols=cols, val=1.0)

        ic = 2

        # Vertical Tail
        if self.num_tails > 0:
            rows = ic + np.arange(self.num_tails)
            cols = np.zeros(self.num_tails)
            self.declare_partials(
                'wetted_areas', Aircraft.VerticalTail.WETTED_AREA,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'fineness_ratios', Aircraft.VerticalTail.FINENESS,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'characteristic_lengths', Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_upper', Aircraft.VerticalTail.LAMINAR_FLOW_UPPER,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_lower', Aircraft.VerticalTail.LAMINAR_FLOW_LOWER,
                rows=rows, cols=cols, val=1.0)
            ic += self.num_tails

        # Fuselage
        if self.num_fuselages > 0:
            rows = ic + np.arange(self.num_fuselages)
            cols = np.zeros(self.num_fuselages)
            self.declare_partials(
                'wetted_areas', Aircraft.Fuselage.WETTED_AREA,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'fineness_ratios', Aircraft.Fuselage.FINENESS,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'characteristic_lengths', Aircraft.Fuselage.CHARACTERISTIC_LENGTH,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_upper', Aircraft.Fuselage.LAMINAR_FLOW_UPPER,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_lower', Aircraft.Fuselage.LAMINAR_FLOW_LOWER,
                rows=rows, cols=cols, val=1.0)
            ic += self.num_fuselages

        # Nacelle
        if self.num_nacelles > 0:
            rows = ic + np.arange(self.num_nacelles)
            cols = np.zeros(self.num_nacelles)
            self.declare_partials(
                'wetted_areas', Aircraft.Nacelle.WETTED_AREA,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'fineness_ratios', Aircraft.Nacelle.FINENESS,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'characteristic_lengths', Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_upper', Aircraft.Nacelle.LAMINAR_FLOW_UPPER,
                rows=rows, cols=cols, val=1.0)
            self.declare_partials(
                'laminar_fractions_lower', Aircraft.Nacelle.LAMINAR_FLOW_LOWER,
                rows=rows, cols=cols, val=1.0)

    def compute(self, inputs, outputs):
        # Wing
        outputs['wetted_areas'][0] = inputs[
            Aircraft.Wing.WETTED_AREA]
        outputs['fineness_ratios'][0] = inputs[
            Aircraft.Wing.FINENESS]
        outputs['characteristic_lengths'][0] = inputs[
            Aircraft.Wing.CHARACTERISTIC_LENGTH]
        outputs['laminar_fractions_upper'][0] = inputs[
            Aircraft.Wing.LAMINAR_FLOW_UPPER]
        outputs['laminar_fractions_lower'][0] = inputs[
            Aircraft.Wing.LAMINAR_FLOW_LOWER]

        # Horizontal Tail
        outputs['wetted_areas'][1] = inputs[
            Aircraft.HorizontalTail.WETTED_AREA]
        outputs['fineness_ratios'][1] = inputs[
            Aircraft.HorizontalTail.FINENESS]
        outputs['characteristic_lengths'][1] = inputs[
            Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH]
        outputs['laminar_fractions_upper'][1] = inputs[
            Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER]
        outputs['laminar_fractions_lower'][1] = inputs[
            Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER]

        ic = 2

        # Vertical Tail
        if self.num_tails > 0:
            for j in range(self.num_tails):
                outputs['wetted_areas'][ic + j] = inputs[
                    Aircraft.VerticalTail.WETTED_AREA]
                outputs['fineness_ratios'][ic + j] = inputs[
                    Aircraft.VerticalTail.FINENESS]
                outputs['characteristic_lengths'][ic + j] = inputs[
                    Aircraft.VerticalTail.CHARACTERISTIC_LENGTH]
                outputs['laminar_fractions_upper'][ic + j] = inputs[
                    Aircraft.VerticalTail.LAMINAR_FLOW_UPPER]
                outputs['laminar_fractions_lower'][ic + j] = inputs[
                    Aircraft.VerticalTail.LAMINAR_FLOW_LOWER]
            ic += self.num_tails

        # Fuselage
        if self.num_fuselages > 0:
            for j in range(self.num_fuselages):
                outputs['wetted_areas'][ic + j] = inputs[Aircraft.Fuselage.WETTED_AREA]
                outputs['fineness_ratios'][ic + j] = inputs[Aircraft.Fuselage.FINENESS]
                outputs['characteristic_lengths'][ic + j] = inputs[
                    Aircraft.Fuselage.CHARACTERISTIC_LENGTH]
                outputs['laminar_fractions_upper'][ic + j] = inputs[
                    Aircraft.Fuselage.LAMINAR_FLOW_UPPER]
                outputs['laminar_fractions_lower'][ic + j] = inputs[
                    Aircraft.Fuselage.LAMINAR_FLOW_LOWER]
            ic += self.num_fuselages

        # Nacelle
        num_engines = self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES)

        wetted_areas = inputs[Aircraft.Nacelle.WETTED_AREA]
        fineness = inputs[Aircraft.Nacelle.FINENESS]
        char_len = inputs[Aircraft.Nacelle.CHARACTERISTIC_LENGTH]
        lam_up = inputs[Aircraft.Nacelle.LAMINAR_FLOW_UPPER]
        lam_low = inputs[Aircraft.Nacelle.LAMINAR_FLOW_LOWER]

        area_list = np.zeros(0, dtype=wetted_areas.dtype)
        fineness_list = np.zeros(0, dtype=fineness.dtype)
        len_list = np.zeros(0, dtype=char_len.dtype)
        lam_up_list = np.zeros(0, dtype=lam_up.dtype)
        lam_low_list = np.zeros(0, dtype=lam_low.dtype)

        for i, num in enumerate(num_engines):
            if num > 0:
                area_list = np.append(area_list, np.tile(wetted_areas[i], num))
                fineness_list = np.append(fineness_list, np.tile(fineness[i], num))
                len_list = np.append(len_list, np.tile(char_len[i], num))
                lam_up_list = np.append(lam_up_list, np.tile(lam_up[i], num))
                lam_low_list = np.append(lam_low_list, np.tile(lam_low[i], num))

        if self.num_nacelles > 0:
            for j in range(self.num_nacelles):
                outputs['wetted_areas'][ic + j] = area_list[j]
                outputs['fineness_ratios'][ic + j] = fineness_list[j]
                outputs['characteristic_lengths'][ic + j] = len_list[j]
                outputs['laminar_fractions_upper'][ic + j] = lam_up_list[j]
                outputs['laminar_fractions_lower'][ic + j] = lam_low_list[j]



'''
Unit test cases for class NamedValues.
'''
import unittest

from aviary.utils.named_values import NamedValues, get_items, get_keys, get_values
from aviary.variable_info.variables import Aircraft, Mission


class NamedValuesTest(unittest.TestCase):
    def test_init(self):
        d = NamedValues()
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues(_empty)
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues(**_empty)
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues(_data1)
        self._do_test_full_equal(d, _data1, [_empty, _data2, _data3])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues(**_data2)
        self._do_test_full_equal(d, _data2, [_empty, _data1, _data3])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues(_data1, **_data2)
        self._do_test_full_equal(d, _data3, [_empty, _data1, _data2])
        a = NamedValues(d)
        self._do_test_full_equal(d, a, ())

    def test_update(self):
        d = NamedValues()
        d.update()
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues()
        a.update(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues()
        d.update(_empty)
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues()
        a.update(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues()
        d.update(**_empty)
        self._do_test_full_equal(d, _empty, [_data1, _data2, _data3])
        a = NamedValues()
        t = tuple(d)
        a.update(t)
        self._do_test_full_equal(d, a, ())

        d = NamedValues()
        d.update(_data1)
        self._do_test_full_equal(d, _data1, [_empty, _data2, _data3])
        a = NamedValues()
        a.update(d)
        self._do_test_full_equal(d, a, ())

        d = NamedValues()
        d.update(**_data2)
        self._do_test_full_equal(d, _data2, [_empty, _data1, _data3])
        a = NamedValues()
        t = tuple(d)
        a.update(t)
        self._do_test_full_equal(d, a, ())

        d = NamedValues()
        d.update(_data1, **_data2)
        self._do_test_full_equal(d, _data3, [_empty, _data1, _data2])
        a = NamedValues()
        t = tuple(d)
        a.update(t)
        self._do_test_full_equal(d, a, ())

    def test_units(self):
        a = NamedValues()
        d = NamedValues(_data1)

        self._do_test_full_equal(d, _data1, (a, _empty, _data2, _data3))

        for key in _data1:
            a.set_val(key, *(d.get_item(key)))

        self._do_test_full_equal(a, d, (_empty, _data2, _data3))
        self._do_test_full_equal(d, _data1, (_empty, _data2, _data3))

        for key in _data1:
            # units support
            val, units = _data1[key]
            aval, aunits = a.get_item(key)

            self.assertEqual(val, aval)
            self.assertEqual(units, aunits)

        self.assertFalse(_nokey in _data1)

        # default (None, None)
        aval, aunits = a.get_item(_nokey)

        self.assertIsNone(aval)
        self.assertIsNone(aunits)

        val = 42
        units = 'min'

        aval, aunits = a.get_item(_nokey, (val, units))

        self.assertEqual(val, aval)
        self.assertEqual(units, aunits)

        key = 'elapsed_time'
        a.set_val(key, val, units)

        val = a.get_val(key, units)

        self.assertEqual(val, aval)

        new_val = 42 * 60  # min -> s

        val = a.get_val(key, 's')

        self.assertEqual(val, new_val)

        # unit conversion is local; item is unchanged
        val, units = a.get_item(key)

        self.assertEqual(val, aval)
        self.assertEqual(units, aunits)

        # still missing: default (None, None)
        aval, aunits = a.get_item(_nokey)

        self.assertIsNone(aval)
        self.assertIsNone(aunits)

        key = _nokey
        val = 314159
        # default units ('unitless')

        a.set_val(key, val)
        aval, aunits = a.get_item(key)

        self.assertEqual(val, aval)
        self.assertEqual(aunits, 'unitless')

    def test_collection(self):
        self.assertNotEqual(len(_data1), 0)
        d = NamedValues(_data1)

        self._do_test_full_equal(d, _data1, (_empty, _data2, _data3))

        # Container
        for key in _data1:
            self.assertTrue(key in d)

        self.assertTrue(_nokey not in d)
        self.assertFalse(_nokey in d)

        # Iterable
        a = tuple(_data1.items())
        b = tuple(iter(d))
        self.assertEqual(a, b)

        b = tuple(d)
        self.assertEqual(a, b)

        b = tuple(get_items(d))
        self.assertEqual(a, b)

        a = tuple(iter(_data1))
        b = tuple(get_keys(d))
        self.assertEqual(a, b)

        a = tuple(_data1.values())
        b = tuple(get_values(d))
        self.assertEqual(a, b)

        # Sized
        self.assertEqual(len(_data1), len(d))

    def test_delete(self):
        # create NamedValues collection a from dictionary _data1.
        a = NamedValues(_data1)
        self._do_test_full_equal(a, _data1, ())

        # remove an item from a.
        a.delete('NUM_FUSELAGES')
        # remove the same item from the dictionary and then make NamedValues collection b
        _data4 = _data1.copy()
        _data4.pop('NUM_FUSELAGES')
        b = NamedValues(_data4)
        # make sure a and b are the same.
        self._do_test_full_equal(a, b, ())

        # make sure it is deleted
        aval, aunits = a.get_item('NUM_FUSELAGES')
        self.assertIsNone(aval)
        self.assertIsNone(aunits)

        # size
        self.assertEqual(len(_data4), len(a))
        self.assertEqual(len(_data4), len(b))

        # delete all
        for key, _ in b:
            a.delete(key)
        self._do_test_full_equal(a, _empty, ())

    def _do_test_full_equal(self, d, eq, ne):
        self.assertEqual(d._mapping, eq)
        self.assertEqual(d, eq)
        self.assertEqual(eq, d)

        for a in ne:
            self.assertNotEqual(d, a)
            self.assertNotEqual(a, d)


_nokey = '## no key ##'

_empty = {}

_data1 = {
    'NUM_FUSELAGES': (1, 'unitless'),
    'NUM_ENGINES': (2, 'unitless'),
    Aircraft.CrewPayload.BAGGAGE_MASS: (7500, 'lbm'),
    Mission.Design.RANGE: (3500, 'NM')
}


def _nounits(a):
    d = {}

    for key, (val, _) in a.items():
        d[key] = val

    return d


_data1_nounits = _nounits(_data1)

_data2 = {
    'NUM_FUSELAGES': (2, 'unitless'),
    'NUM_ENGINES': (4, 'unitless'),
    'PASSENGER_MASS': (200, 'lbm'),
}

# _data2_nounits = _nounits(_data2)

_data3 = dict(_data1, **_data2)
# _data3_nounits = _nounits(_data3)


if __name__ == '__main__':
    unittest.main()



import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class CLmaxCalculation(om.ExplicitComponent):
    def setup(self):

        # inputs

        # from component 1 outputs
        self.add_input(
            "VLAM8",
            val=0.74444322,
            units='unitless',
            desc="VLAM8: sensitivity of flap clean wing maximum lift coefficient to wing sweep angle",
        )
        self.add_input(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=1118.21948771,
            units="ft/s",
            desc="SA: speed of sound at sea level",
        )

        # from component 3 outputs
        add_aviary_input(self, Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM, val=1.5)
        self.add_input(
            "VLAM1",
            val=0.97217,
            units='unitless',
            desc="VLAM1: sensitivity of clean wing maximum lift coefficient to wing aspect ratio",
        )
        self.add_input(
            "VLAM2",
            val=1.09948,
            units='unitless',
            desc="VLAM2: sensitivity of clean wing maximum lift coefficient to wing thickness to chord ratio",
        )
        self.add_input(
            "VLAM3",
            val=0.97217,
            units='unitless',
            desc="VLAM3: sensitivity of flap clean wing maximum lift coefficient to aspect ratio",
        )
        self.add_input(
            "VLAM4",
            val=1.25725,
            units='unitless',
            desc="VLAM4: sensitivity of flap clean wing maximum lift coefficient slope to wing thickness",
        )
        self.add_input(
            "VLAM5",
            val=1.0,
            units='unitless',
            desc="VLAM5: sensitivity of flap clean wing maximum lift coefficient to wing flap to chord ratio",
        )
        self.add_input(
            "VLAM6",
            val=1.0,
            units='unitless',
            desc="VLAM6: sensitivity of flap clean wing maximum lift coefficient to wing flap deflection",
        )
        self.add_input(
            "VLAM7",
            val=0.735,
            units='unitless',
            desc="VLAM7: sensitivity of flap clean wing maximum lift coefficient to wing flap span",
        )
        self.add_input(
            "VLAM13", val=1.03512, units='unitless', desc="VLAM13: reynolds number correction factor"
        )
        self.add_input(
            "VLAM14", val=0.99124, units='unitless', desc="VLAM14: mach number correction factor "
        )

        # other inputs

        add_aviary_input(self, Aircraft.Wing.LOADING, val=128)

        self.add_input(
            Dynamic.Mission.STATIC_PRESSURE, val=(14.696 * 144), units="lbf/ft**2", desc="P0: static pressure"
        )

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=12.61)

        add_aviary_input(self, Aircraft.Wing.MAX_LIFT_REF, val=1.15)
        add_aviary_input(self, Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM, val=0.93)
        self.add_input(
            "VLAM9",
            val=0.9975,
            units='unitless',
            desc="VLAM9: sensitivity of slat clean wing maximum lift coefficient to slat chord",
        )
        self.add_input(
            "VLAM10",
            val=0.74,
            units='unitless',
            desc="VLAM10: sensitivity of slat clean wing maximum lift coefficient to slat deflection angle",
        )
        self.add_input(
            "VLAM11",
            val=0.84232,
            units='unitless',
            desc="VLAM11: sensitivity of slat clean wing mazimum lift coefficient to slat span",
        )
        self.add_input(
            "VLAM12",
            val=0.79208,
            units='unitless',
            desc="VLAM12: sensitivity of slat clean wing maximum lift coefficient to leading edge sweepback",
        )
        self.add_input("fus_lift", val=0.05498, units='unitless',
                       desc="DELCLF: fuselage lift increment")
        self.add_input(
            "kinematic_viscosity",
            val=0.15723e-03,
            units="ft**2/s",
            desc="XKV: kinematic viscosity",
        )
        self.add_input(
            Dynamic.Mission.TEMPERATURE,
            val=518.67,
            units="degR",
            desc="T0: static temperature of air cross wing",
        )

        # outputs

        self.add_output("CL_max", val=2.8155,
                        desc="CLMAX: maximum lift coefficient", units="unitless")
        self.add_output(Dynamic.Mission.MACH, val=0.17522,
                        units='unitless', desc="SMN: mach number")
        self.add_output("reynolds", val=157.1111, units='unitless',
                        desc="RNW: reynolds number")

    def setup_partials(self):

        self.declare_partials(
            "CL_max",
            [
                Aircraft.Wing.MAX_LIFT_REF,
                "VLAM1",
                "VLAM2",
                Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM,
                "VLAM9",
                "VLAM10",
                "VLAM11",
                "VLAM12",
                "VLAM13",
                "VLAM14",
                "fus_lift",
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            Dynamic.Mission.MACH,
            [
                Aircraft.Wing.LOADING,
                Dynamic.Mission.STATIC_PRESSURE,
                Aircraft.Wing.MAX_LIFT_REF,
                "VLAM1",
                "VLAM2",
                Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM,
                "VLAM9",
                "VLAM10",
                "VLAM11",
                "VLAM12",
                "VLAM13",
                "VLAM14",
                "fus_lift",
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "reynolds",
            [
                "kinematic_viscosity",
                Dynamic.Mission.SPEED_OF_SOUND,
                Aircraft.Wing.AVERAGE_CHORD,
                Dynamic.Mission.STATIC_PRESSURE,
                Aircraft.Wing.LOADING,
                Aircraft.Wing.MAX_LIFT_REF,
                "VLAM1",
                "VLAM2",
                Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM,
                "VLAM9",
                "VLAM10",
                "VLAM11",
                "VLAM12",
                "VLAM13",
                "VLAM14",
                "fus_lift",
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )

    def compute(self, inputs, outputs):

        VLAM1 = inputs["VLAM1"]
        VLAM2 = inputs["VLAM2"]
        VLAM3 = inputs["VLAM3"]
        VLAM4 = inputs["VLAM4"]
        VLAM5 = inputs["VLAM5"]
        VLAM6 = inputs["VLAM6"]
        VLAM7 = inputs["VLAM7"]
        VLAM8 = inputs["VLAM8"]
        VLAM9 = inputs["VLAM9"]
        VLAM10 = inputs["VLAM10"]
        VLAM11 = inputs["VLAM11"]
        VLAM12 = inputs["VLAM12"]
        VLAM13 = inputs["VLAM13"]
        VLAM14 = inputs["VLAM14"]

        sos = inputs[Dynamic.Mission.SPEED_OF_SOUND]
        wing_loading = inputs[Aircraft.Wing.LOADING]
        P = inputs[Dynamic.Mission.STATIC_PRESSURE]
        avg_chord = inputs[Aircraft.Wing.AVERAGE_CHORD]
        kinematic_viscosity = inputs["kinematic_viscosity"]
        max_lift_reference = inputs[Aircraft.Wing.MAX_LIFT_REF]
        leading_lift_increment = inputs[Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM]
        fus_lift = inputs["fus_lift"]
        trailing_lift_increment = inputs[Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM]

        # outputs

        outputs["CL_max"] = CL_max = (
            max_lift_reference * VLAM1 * VLAM2
            + trailing_lift_increment * VLAM3 * VLAM4 * VLAM5 * VLAM6 * VLAM7 * VLAM8
            + leading_lift_increment * VLAM9 * VLAM10 * VLAM11 * VLAM12
        ) * VLAM13 * VLAM14 + fus_lift

        Q1 = wing_loading / CL_max

        outputs[Dynamic.Mission.MACH] = mach = (Q1 / 0.7 / P) ** 0.5

        VK = mach * sos
        outputs["reynolds"] = reynolds = (avg_chord * VK / kinematic_viscosity) / 100000



import numpy as np
import openmdao.api as om
from pathlib import Path
import csv
import pkg_resources
import os

from aviary.utils.aviary_values import AviaryValues, get_keys
from aviary.variable_info.enums import ProblemType, EquationsOfMotion, LegacyCode
from aviary.variable_info.functions import add_aviary_output, add_aviary_input
from aviary.variable_info.variable_meta_data import _MetaData


class Null:
    '''
    This can be used to divert outputs, such as stdout, to improve performance
    '''

    def write(self, *args, **kwargs):
        pass

    def flush(self, *args, **kwargs):
        pass


def set_aviary_initial_values(model, inputs, meta_data=_MetaData):
    '''
    This function sorts through all the input
    variables to an Aviary model, and for those
    which are not options it sets the input
    value to be the value in the inputs, or
    to be the default if the value is not in the
    inputs.

    In the case when the value is not input nor
    present in the default, nothing is set.
    '''
    for key in meta_data:
        if ':' not in key or key.startswith('dynamic:'):
            continue
        if not meta_data[key]['option']:
            if key in inputs:
                val, units = inputs.get_item(key)
            else:
                val = meta_data[key]['default_value']
                units = meta_data[key]['units']

                if val is None:
                    # optional, but no default value
                    continue

            model.set_input_defaults(key, val=val, units=units)


def apply_all_values(aircraft_values: AviaryValues, prob):
    for var_name in get_keys(aircraft_values):
        var_data, var_units = aircraft_values.get_item(var_name)
        try:
            prob.set_val(var_name, val=var_data, units=var_units)
        except KeyError:
            pass
    return prob


def convert_strings_to_data(string_list):
    # convert_strings_to_data will convert a list of strings to usable data.
    # Strings that can't be converted to numbers will attempt to store as a logical,
    # otherwise they are passed as is
    value_list = [0]*len(string_list)
    for ii, dat in enumerate(string_list):
        dat = dat.strip('[]')
        try:
            # if the value is a number store it as a float or an int as appropriate
            # BUG this returns floats that can be converted to int (e.g. 1.0) as an int (1), even if the variable requires floats
            value_list[ii] = int(float(dat)) if float(
                dat).is_integer() else float(dat)
        except ValueError:
            # store value as a logical if it is a string that represents True or False
            if dat.lower() == 'true':
                value_list[ii] = True
            elif dat.lower() == 'false':
                value_list[ii] = False
            else:
                # if the value isn't a number or a logial, store it as a string
                value_list[ii] = dat
        except Exception as e:
            print('Exception', e)
    return value_list


def set_value(var_name, var_value, aviary_values: AviaryValues, units=None, is_array=False, meta_data=_MetaData):
    if var_name in aviary_values:
        current_value, current_units = aviary_values.get_item(var_name)
    else:
        current_value = meta_data[var_name]['default_value']
        current_units = meta_data[var_name]['units']

    if units == None:
        if current_units:
            units = current_units
        else:
            units = meta_data[var_name]['units']
        #     raise ValueError("You have specified a new variable without any units")

    if is_array:
        var_value = np.atleast_1d(var_value)
    elif len(var_value) == 1 and not isinstance(current_value, list):
        # if only a single value is provided, don't store it as a list
        var_value = var_value[0]

    # TODO handle enums in an automated method via checking metadata for enum type
    if var_name == 'problem_type':
        var_values = ProblemType[var_value]
    if var_name == 'settings:equations_of_motion':
        var_values = EquationsOfMotion(var_value)
    if var_name == 'settings:mass_method':
        var_values = LegacyCode(var_value)

    aviary_values.set_val(var_name, val=var_value, units=units, meta_data=meta_data)
    return aviary_values


def create_opts2vals(all_options: list, output_units: dict = {}):
    """
    create_opts2vals creates a component that converts options to outputs.

    Parameters
    ----------
    all_options : list of strings
        Each string is the name of an option in aviary_options.
    output_units : dict of units, optional
        This optional input allows the user to specify the units that will be used while
        adding the outputs. Only the outputs that shouldn't use their default units need
        to be specified. Each key should match one of the names in all_options, and each
        value must be a string representing a valid unit in openMDAO.

    Returns
    -------
    OptionsToValues : ExplicitComponent
        An explicit component that takes in an AviaryValues object that contains any
        options that need to be converted to outputs. There are no inputs to this
        component, only outputs. If the resulting component is added directly to a
        Group, the output variables will have the same name as the options they
        represent. If you need to rename them to prevent conflicting names in the
        group, running add_opts2vals will add the prefix "option:" to the name.
    """

    def configure_output(option_name: str, aviary_options: AviaryValues):
        option_data = aviary_options.get_item(option_name)
        out_units = output_units[option_name] if option_name in output_units.keys(
        ) else option_data[1]
        return {'val': option_data[0], 'units': out_units}

    class OptionsToValues(om.ExplicitComponent):
        def initialize(self):
            self.options.declare(
                'aviary_options', types=AviaryValues,
                desc='collection of Aircraft/Mission specific options'
            )

        def setup(self):
            for option_name in all_options:
                output_data = configure_output(
                    option_name, self.options['aviary_options'])
                add_aviary_output(self, option_name,
                                  val=output_data['val'], units=output_data['units'])

        def compute(self, inputs, outputs):
            aviary_options: AviaryValues = self.options['aviary_options']
            for option_name in all_options:
                output_data = configure_output(option_name, aviary_options)
                outputs[option_name] = aviary_options.get_val(
                    option_name, units=output_data['units'])

    return OptionsToValues


def add_opts2vals(Group: om.Group, OptionsToValues, aviary_options: AviaryValues):
    """
    Add the OptionsToValues component to the specified Group.

    Parameters
    ----------
    Group : Group
        The group or model the component should be added to.
    OptionsToValues : ExplicitComponent
        This is the explicit component that was created by create_opts2vals.
    aviary_options : AviaryValues
        aviary_options is an AviaryValues object that contains all of the options
        that need to be converted to outputs.

    Returns
    -------
    Opts2Vals : Group
        A group that wraps the OptionsToValues component in order to rename its
        variables with a prefix to keep them separate from any similarly named
        variables in the original group the component is being added to.
    """

    class Opts2Vals(om.Group):
        def initialize(self):
            self.options.declare(
                'aviary_options', types=AviaryValues,
                desc='collection of Aircraft/Mission specific options'
            )

        def setup(self):
            self.add_subsystem('options_to_values', OptionsToValues(
                aviary_options=aviary_options))

        def configure(self):
            all_output_data = self.options_to_values.list_outputs(out_stream=None)
            list_of_outputs = [(name, 'option:'+name) for name, data in all_output_data]
            self.promotes('options_to_values', list_of_outputs)

    Group.add_subsystem('opts2vals', Opts2Vals(
        aviary_options=aviary_options),
        promotes_outputs=['*'])

    return Group


def create_printcomp(all_inputs: list, input_units: dict = {}, meta_data=_MetaData):
    """
    Creates a component that prints the value of all inputs.

    Parameters
    ----------
    all_inputs : list of strings
        Each string is the name of a variable in the system
    input_units : dict of units, optional
        This optional input allows the user to specify the units that will be used while
        adding the inputs. Only the variables that shouldn't use their default units need
        to be specified. Each key should match one of the names in all_inputs, and each
        value must be a string representing a valid unit in openMDAO.

    Returns
    -------
    PrintComp : ExplicitComponent
        An explicit component that can be added to a group to print the current values
        of connected variables. There are no outputs from this component, only inputs.
    """

    def get_units(variable_name):
        if variable_name in input_units.keys():
            return input_units[variable_name]
        elif variable_name in meta_data:
            return meta_data[variable_name]['units']
        else:
            return None

    class PrintComp(om.ExplicitComponent):

        def setup(self):
            for variable_name in all_inputs:
                units = get_units(variable_name)
                if ':' in variable_name:
                    add_aviary_input(self, variable_name, units=units)
                else:
                    self.add_input(variable_name, units=units)

        def compute(self, inputs, outputs):
            print_string = ['v'*20]
            for variable_name in all_inputs:
                units = get_units(variable_name)
                print_string.append('{} {} {}'.format(
                    variable_name, inputs[variable_name], units))
            print_string.append('^'*20)
            print('\n'.join(print_string))

    return PrintComp


def promote_aircraft_and_mission_vars(group):
    external_outputs = []
    for comp in group.system_iter(recurse=False):

        # Skip all aviary systems.
        if comp.name == 'core_subsystems':
            continue

        out_names = [item for item in comp._var_allprocs_prom2abs_list['output']]
        in_names = [item for item in comp._var_allprocs_prom2abs_list['input']]

        external_outputs.extend(out_names)

        # Locally promote aircraft:* and mission:* only.
        promote_in = []
        promote_out = []

        for stem in ['mission:', 'aircraft:', 'dynamic:']:
            for name in out_names:
                if name.startswith(stem):
                    promote_out.append(f'{stem}*')
                    break

            for name in in_names:
                if name.startswith(stem):
                    promote_in.append(f'{stem}*')
                    break

        group.promotes(comp.name, outputs=promote_out, inputs=promote_in)

    return external_outputs


def get_path(path: [str, Path], verbose: bool = False) -> Path:
    """
    Convert a string or Path object to an absolute Path object, prioritizing different locations.

    This function attempts to find the existence of a path in the following order:
    1. As an absolute path.
    2. Relative to the current working directory.
    3. Relative to the Aviary package.

    If the path cannot be found in any of the locations, a FileNotFoundError is raised.

    Parameters
    ----------
    path : str or Path
        The input path, either as a string or a Path object.
    verbose : bool, optional
        If True, prints the final path being used. Default is False.

    Returns
    -------
    Path
        The absolute path to the file.

    Raises
    ------
    FileNotFoundError
        If the path is not found in any of the prioritized locations.
    """

    # Store the original path for reference in error messages.
    original_path = path

    # If the input is a string, convert it to a Path object.
    if isinstance(path, str):
        path = Path(path)

    # Check if the path exists as an absolute path.
    if not path.exists():
        # If not, try finding the path relative to the current working directory.
        path = Path.cwd() / path

    # If the path still doesn't exist, attempt to find it relative to the Aviary package.
    if not path.exists():
        # Determine the path relative to the Aviary package.
        aviary_based_path = Path(
            pkg_resources.resource_filename('aviary', original_path))
        if verbose:
            print(
                f"Unable to locate '{original_path}' as an absolute or relative path. Trying Aviary package path: {aviary_based_path}")
        path = aviary_based_path

    # If the path still doesn't exist in any of the prioritized locations, raise an error.
    if not path.exists():
        raise FileNotFoundError(
            f'File not found in absolute path: {original_path}, relative path:{Path.cwd() / path}, or Aviary-based path: {Path(pkg_resources.resource_filename("aviary", original_path))}'
        )

    # If verbose is True, print the path being used.
    if verbose:
        print(f'Using {path} for file.')

    return path
